/* pingpong - measure effective bandwidth and latency */

#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <errno.h>
#include "getus.h"
#include "mpi.h"

#define MAXSIZE (2*1024*1024)
#define MINSIZE (0)
#define INCSIZE (2)
#define INCOP   *=
#define REPEAT  (1000)
#define AVERAGE (5) 

#define CALIBRATION_LOOPS 200

#define TAG_PING 1
#define TAG_PONG 2

//#define __ONE_WAY__
#define MASTER_NODE  0
#define SLAVE_NODE   3
#define GATEWAY      2
#define MASTER_NODE2 1
#define SLAVE_NODE2  4

#define COLOR1   101
#define COLOR2   303
#define NO_COLOR MPI_UNDEFINED

#ifdef  __ONE_WAY__
#define CORREC      1
#else
#define CORREC      2
#endif

/* define DETAIL if you want to create histogramms by measuring
   the latency of each single ping-pong transfer */
/*#if 1
#define DETAIL
#endif
*/

#ifdef linux
#define longlong_t long long
#endif

char *buffer;
char *exename;
int min_size, max_size, inc_size, repeats, average;
int myrank, mysize;
static FILE* fpGlobal  = NULL;
static FILE* fpGlobal2 = NULL;
#ifdef DETAIL
static FILE* fpDetail  = NULL;
#endif

void ping (int to, int from, FILE* file, MPI_Comm com);
void pong (int , MPI_Comm com);
void do_nothing(void);

int  main(int argc, char **argv) 
{
  //MPI_Status status;
   int  color;
   int  key;
   char fname[128], fname2[128];
   MPI_Comm Newcomm;
   
   MPI_Init(&argc, &argv);
   MPI_Comm_rank(MPI_COMM_WORLD, &myrank);
   MPI_Comm_size(MPI_COMM_WORLD, &mysize);

   if ((myrank == MASTER_NODE))
     {
	color = COLOR1;
	key   = 0;
     }
   else if ( myrank == SLAVE_NODE)
     {
	color = COLOR1;
	key   = 1;
     }
   else  if (myrank == MASTER_NODE2) 
     {
	color = COLOR2;
	key = 0;
     }
   else if ( myrank == SLAVE_NODE2)
     {
	color = COLOR2;
	key = 1;
     }
   else
     {
	color = NO_COLOR;
	key = 2;
     }
   
   MPI_Comm_split(MPI_COMM_WORLD , color, key, &Newcomm);
   
   if (myrank == MASTER_NODE )
     {
	strcpy(fname, argv[0]);
	strcat(fname,".global");
	fpGlobal = fopen(fname,"w");	
     }
   
   if (myrank == MASTER_NODE2 )
     {
	strcpy(fname2, argv[0]);
	strcat(fname2,".global2");
	fpGlobal2 = fopen(fname2,"w");	
     }
   exename = argv[0];

   fprintf(stdout, "<<<< MYRANK IS : %i >>>>\n", myrank);
   fflush(stdout);
   
   /* set run parameters */
   if (argc != 5) 
     {
	printf ("usage: pingpong min_size max_size repeats average\n");
	printf ("using default values for this run\n");
	
	min_size = MINSIZE;
	max_size = MAXSIZE;
	inc_size = INCSIZE;
	repeats  = REPEAT;
	average  = AVERAGE ;
     } 
   else 
     {
	min_size = atoi( argv[1] );
	max_size = atoi( argv[2] );
	inc_size = INCSIZE;
	repeats  = atoi( argv[3] );
	average  = atoi( argv[4] );
     }
   buffer = (char *)malloc (max_size);
   
   if (myrank == MASTER_NODE)
     {
	ping( SLAVE_NODE , myrank, fpGlobal, Newcomm);
	MPI_Barrier(MPI_COMM_WORLD);
     }
   else if (myrank == MASTER_NODE2)
     {
	ping( SLAVE_NODE2 , myrank, fpGlobal2, Newcomm);
	MPI_Barrier(MPI_COMM_WORLD);
     }
   else if (myrank == SLAVE_NODE )
     {
	pong( MASTER_NODE , Newcomm);
	MPI_Barrier(MPI_COMM_WORLD);
     }
   else if (myrank == SLAVE_NODE2 )
     {
	pong( MASTER_NODE2 , Newcomm);
	MPI_Barrier(MPI_COMM_WORLD);
     }
   else if (myrank == GATEWAY)
     {
       //do_nothing();
       MPI_Barrier(MPI_COMM_WORLD);
     }
   
   if (myrank ==  MASTER_NODE ) {
      fclose(fpGlobal);
   }
   if (myrank ==  MASTER_NODE2 ) {
      fclose(fpGlobal2);
   }
   
   free (buffer);
   MPI_Finalize();
   return 0; 
}


void ping( int to, int from, FILE* file, MPI_Comm com) 
{
   MPI_Status status;
   double starttime, totaltime;
   double averagelat = 0, averagebw = 0;
   
#ifdef DETAIL
   char fname[128];
   char bytes[128];
   double getticks_overhead; 
   longlong_t hr_start, hr_end;
   longlong_t calibration = 0;
   longlong_t *timings;
#endif
   
   int i, j, k;
   int firstrun = 1;
   
#ifdef DETAIL
   fprintf (stderr, "Calibrating...");
   for (i = 0; i < CALIBRATION_LOOPS; i++) {
      GETTICKS(&hr_start);
      GETTICKS(&hr_end);
      calibration += hr_end - hr_start;
   }
   getticks_overhead = ((double)calibration)/(CALIBRATION_LOOPS); 
   fprintf (stderr, "gethrtime() overhead is %6.3f\n", getticks_overhead);
   timings = (longlong_t *)malloc (repeats*sizeof(longlong_t));
#endif
   
   printf("pingpong from %d to %d\n\n", from, to);
   fprintf(file, "# MsgSize[byte]  Repeats  Average Bandwidth[MB/s]  Latency[us]\n");
   fflush(stdout);
   
   for( i = min_size; i <= max_size; i INCOP inc_size) 
     {
	averagelat = 0 ; 
	averagebw  = 0;
	
	if ((!firstrun) && (i == 0)) 
	  {
	     i++;
	     if (i > max_size)
	       break;
	  }
	firstrun = 0;     
	
	for( k = 0 ; k < average; k++) 
	  {
	     MPI_Barrier(com);
	     
#ifdef DETAIL
	     strcpy (fname, exename);
	     sprintf (bytes, "_id%d_%db.detail", myrank, i);
	     strcat (fname, bytes);
	     fpDetail = fopen (fname,"w");
	     fprintf (fpDetail, "# ticks for ping-pong, msgsize = %d byte\n# getticks_overhead = %6.3f \n", 
		      i, getticks_overhead);
#endif
	     starttime = MPI_Wtime();
	     for( j = 0; j < repeats; j++) 
	       {
#ifdef DETAIL
		  GETTICKS(&hr_start);
#endif
		  MPI_Send( buffer, i, MPI_CHAR, to, TAG_PING, MPI_COMM_WORLD);
		  
#ifdef __ONE_WAY__	   
		  if (j == repeats - 1)
#endif 	     
		    MPI_Recv( buffer, i, MPI_CHAR, to, TAG_PONG, MPI_COMM_WORLD, &status);        
		  
#ifdef DETAIL
		  GETTICKS(&hr_end);
		  timings[j] = hr_end - hr_start;
#endif
	       }
	     totaltime   = MPI_Wtime() - starttime ;	
	     averagelat += (totaltime*1e+6)/((double)(repeats*(CORREC)));
	     averagebw  += (((double)i)*(CORREC)*repeats)/totaltime ;
	  }
	
	printf("size = %7d,  bandwidth = %7.3lf MB/s latency = %7.3lf usec.,\n",
	       i, ((averagebw/((double)(1024*1024)))/((double)average)), averagelat/((double)average));
	
	fprintf(file,"%7d \t %7.3f \t %7.3f\n", i,
		((averagebw/((double)(1024*1024)))/((double)average)), averagelat/((double)average));
	
	fflush(stdout);
	fflush(file);   
	
#ifdef DETAIL
	for( j = 0; j < repeats; j++) {
	   fprintf (fpDetail, "%d \t %6.3f \n", j, (double)timings[j]);
	}
	fclose(fpDetail);
#endif
	
     }
}


void pong( int to , MPI_Comm com)
{
   MPI_Status status;
   int i,j,k;
   int firstrun = 1;
   
   
   for( i = min_size; i <= max_size; i INCOP inc_size) 
     {
	if ((!firstrun) && (i == 0)) 
	  {
	     i++;
	     if (i > max_size)
	       break;
	  }
	firstrun = 0;
	
	for( k = 0 ; k < average; k++) 
	  {
	     MPI_Barrier(com);
	     
	     for(j = 0; j < repeats; j++) 
	       {
		  MPI_Recv( buffer, i, MPI_CHAR, to, TAG_PING, MPI_COMM_WORLD, &status);
		  
#ifdef __ONE_WAY__
		  if ( j == repeats - 1)
#endif 
		    MPI_Send( buffer, i, MPI_CHAR, to, TAG_PONG, MPI_COMM_WORLD);
	       }
	  }  
     }
}

void do_nothing( void )
{
   int i , j;
   
   for( i = min_size; i <= max_size; i INCOP inc_size) 
     {
	for( j = 0 ; j < average; j++) 
	  {
	     MPI_Barrier(MPI_COMM_WORLD);
	  }
     }
}


