/*
 *  $Id: adi2cancel.c,v 1.1.1.1 2003/07/30 15:35:47 lprylli Exp $
 *
 *  (C) 1996 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */

/*#include "mpid.h"*/
/*#include "mpiddev.h"*/
/*#include "mpimem.h"*/
#include "hot_stuff.h"
#include "queue.h"

/*
 * This file contains the routines to handle canceling a message
 *
 * Note: for now, cancel will probably only work on unmatched receives.
 * However, this code provides the hooks for supporting more complete
 * cancel implementations.
 */

void MPID_SendCancel( MPI_Request request, int *error_code )
{  
	/* begin MPID_SendCancel */

    MPIR_SHANDLE *shandle = &request->shandle;

    MPID_HANDLE_MUTEX_LOCK( shandle );
    shandle->is_cancelled = 0;
    MPID_HANDLE_MUTEX_UNLOCK( shandle );

    MPID_CH_Mad_SendCancelPkt(&request, error_code); 
    
    if (*error_code == MPI_SUCCESS) 
      {  
	MPID_HANDLE_WAIT4CANCEL( shandle );	
	MPID_HANDLE_MUTEX_LOCK( shandle );
	if (shandle->is_cancelled) 
	  { 
	    if (shandle->handle_type == MPIR_PERSISTENT_SEND) 
	      {
		MPIR_PSHANDLE *pshandle = (MPIR_PSHANDLE *)request;
		pshandle->active = 0; 
	      }
	  }
	MPID_HANDLE_MUTEX_UNLOCK( shandle );
      }  
    
} 

 
void MPID_RecvCancel( MPI_Request request, int *error_code )
{
    MPIR_RHANDLE *rhandle = &request->rhandle;

    DEBUG_PRINT_MSG("S Starting RecvCancel"); 

    MPID_HANDLE_MUTEX_LOCK( rhandle );
    /* First, try to find in pending receives */
    if (MPID_Dequeue( &MPID_recvs.posted, rhandle ) == 0) 
      {
	/* Mark the request as cancelled */
	rhandle->s.MPI_TAG = MPIR_MSG_CANCELLED;
	
	/* Mark it as complete */
	MPID_HANDLE_COMPLETE_READY( rhandle );	
	
	/* Note that the request is still active until we complete it with
	   a wait/test operation */
      }
    if (rhandle->handle_type == MPIR_PERSISTENT_RECV)
      {
	MPIR_PRHANDLE *prhandle = (MPIR_PRHANDLE *)request;
	prhandle->active = 0; 
      }
    
    MPID_HANDLE_MUTEX_UNLOCK( rhandle );
    /* Mark the request as not cancelled */
    /* tag is already set >= 0 as part of receive */
    /* rhandle->s.tag = 0; */
    /* What to do about an inactive persistent receive? */
    
    /* In the case of a partly completed rendezvous receive, we might
       want to do something */
    *error_code = MPI_SUCCESS;
    DEBUG_PRINT_MSG("E Exiting RecvCancel");
}
