/*
 *
 *
 *  (C) 1995 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */

/*  #include "mpid.h" */
/*  #include "mpiddev.h" */
#include "hot_stuff.h"
/* flow.h includs the optional flow control for eager delivery */
#include "flow.h"


/***************************************************************************/
/*
 * Multi-protocol, Multi-device support for 2nd generation ADI.
 * We start with support for blocking, contiguous sends.
 * Note the 'msgrep' field; this gives a hook for heterogeneous systems
 * which can be ignored on homogeneous systems.
 */
/***************************************************************************/
void MPID_SendContig( struct MPIR_COMMUNICATOR *comm_ptr, 
		      void                     *buf, 
		      int                      len, 
		      int                      src_lrank, 
		      int                      tag, 
		      int                      context_id, 
		      int                      dest_grank, 
		      MPID_Msgrep_t            msgrep, 
		      int                      *error_code )
{
  
  MPID_Device     *dev    = MPID_devset->dev[dest_grank];
  p_mad_channel_t channel = MPID_Get_Channel_From_Communicator( comm_ptr );
  int             threshold;
  int             (*fcn) ( void *, 
			   int, 
			   int, 
			   int, 
			   int, 
			   int, 
			   MPID_Msgrep_t,
			   MPIR_SHANDLE *,
			   p_mad_channel_t );
  
  /* The one error test that makes sense here */
  if (buf == 0 && len > 0) {
    *error_code = MPI_ERR_BUFFER;
    return;
  }
  
  if ( dest_grank == MPID_MyWorldRank )
    { 
      threshold = dev->long_len;
    }
  else
    {
      threshold = Threshold_From_Channel( channel , dev->channel_table );
    }
  
  /* Choose the function based on the message length in bytes */
  if (len < threshold)
    {
      fcn = dev->short_msg->send;
    }
/*  
 else if (len < dev->vlong_len && MPID_FLOW_MEM_OK(len , dest_grank)) 
 {
      fcn = dev->long_msg->send;
 }
 */
   else
     {
	fcn = dev->vlong_msg->send;
     }
    
  DEBUG_TEST_FCN(fcn,"dev->proto->send");
  *error_code = (*(fcn))( buf, 
			  len, 
			  src_lrank, 
			  tag, 
			  context_id, 
			  dest_grank, 
			  msgrep,
			  NULL,
			  channel );
}

void MPID_IsendContig( struct MPIR_COMMUNICATOR *comm_ptr, 
		       void                     *buf, 
		       int                      len, 
		       int                      src_lrank, 
		       int                      tag, 
		       int                      context_id, 
		       int                      dest_grank, 
		       MPID_Msgrep_t            msgrep, 
		       MPI_Request              request, 
		       int                      *error_code )
{
    MPID_Device     *dev    = MPID_devset->dev[dest_grank];
    p_mad_channel_t channel = MPID_Get_Channel_From_Communicator( comm_ptr );
    int             threshold;
    int             (*fcn) ( void *, 
			     int, 
			     int, 
			     int, 
			     int, 
			     int, 
			     MPID_Msgrep_t, 
			     MPIR_SHANDLE *,
			     p_mad_channel_t );
    
    /* The one error test that makes sense here */
    if (buf == NULL && len > 0) {
	*error_code = MPI_ERR_BUFFER;
	return;
    }

    /* Just in case; make sure that finish is NULL */
    MPID_HANDLE_MUTEX_LOCK( &(request->shandle) );
    (request->shandle).finish  = NULL;
    MPID_HANDLE_MUTEX_UNLOCK( &(request->shandle) );

    
    if ( dest_grank == MPID_MyWorldRank )
      { 
	threshold = dev->long_len;
      }
    else
      {
	threshold = Threshold_From_Channel( channel , dev->channel_table );
      }
    
    /* Choose the function based on the message length in bytes */
    if (len < threshold)
      {
	fcn = dev->short_msg->isend;
      }
/*    else if (len < dev->vlong_len && MPID_FLOW_MEM_OK(len,dest_grank)) 
      {
	fcn = dev->long_msg->isend;
      }
 */
    else
      {
	fcn = dev->vlong_msg->isend;
      }
    
    DEBUG_TEST_FCN(fcn,"dev->proto->isend");
    
    *error_code = (*(fcn))( buf, 
			    len, 
			    src_lrank, 
			    tag, 
			    context_id, 
			    dest_grank, 
			    msgrep, 
			    &(request->shandle),
			    channel );
}

/* Bsend is just a test for short send */
void MPID_BsendContig( struct MPIR_COMMUNICATOR *comm_ptr, 
		       void                     *buf, 
		       int                      len, 
		       int                      src_lrank, 
		       int                      tag, 
		       int                      context_id, 
		       int                      dest_grank, 
		       MPID_Msgrep_t            msgrep, 
		       int                      *error_code )
{
  MPID_Device      *dev    = MPID_devset->dev[dest_grank];
  p_mad_channel_t  channel = MPID_Get_Channel_From_Communicator( comm_ptr );
  int              threshold;
  int              rc;
    
  if ( dest_grank == MPID_MyWorldRank )
    { 
      threshold = dev->long_len;
    }
  else
    {
      threshold = Threshold_From_Channel( channel , dev->channel_table );
    }
  
  if (len < threshold) 
    {
      DEBUG_TEST_FCN(dev->short_msg->send,"dev->short->send");
      rc = (*dev->short_msg->send)( buf, 
				    len, 
				    src_lrank, 
				    tag, 
				    context_id, 
				    dest_grank, 
				    msgrep,
				    NULL,
				    channel );
    }
  else
    {
      rc = MPIR_ERR_MAY_BLOCK;
    }
  
  *error_code = rc;
}
/*-----------------------------------------------------------*/
int MPID_SendIcomplete( 
	MPI_Request request,
	int         *error_code)
{
  MPIR_SHANDLE *shandle = &(request->shandle);
  
  if(MPID_HANDLE_TEST_4COMPLETE( shandle ))
    { 
      return 1; 
    }
  
  if (shandle->test)
    { 
      *error_code = (*(shandle->test))( shandle );
    }
  return 0; 
}
/*------------------------------------------------------------*/
void MPID_SendComplete( MPI_Request request,
			int         *error_code )
{
  MPIR_SHANDLE *shandle = &(request->shandle);
    
  MPID_HANDLE_MUTEX_LOCK( shandle);
  MPID_HANDLE_MUTEX_LOCK_4COMPLETE( shandle );
  if(!(MPID_HANDLE_TEST_4COMPLETE( shandle )))
    {
      DEBUG_PRINT_MSG( "Entering while !shandle->is_complete" );
      MPID_HANDLE_MUTEX_UNLOCK_4COMPLETE( shandle );
      if( shandle->wait )
	{
	  MPID_HANDLE_MUTEX_UNLOCK( shandle);
	  *error_code = (*(shandle->wait))( shandle );
	} 
      else 
	{ 
	  MPID_HANDLE_MUTEX_UNLOCK( shandle);
	  *error_code = MPID_WaitForCompleteSend( shandle );
	}
      DEBUG_PRINT_MSG( "Leaving while !shandle->is_complete" );
    }
  else
    {
      MPID_HANDLE_MUTEX_UNLOCK_4COMPLETE( shandle );
      MPID_HANDLE_MUTEX_UNLOCK( shandle);
    }
}





