
/*******************************************************************************/
#include "hot_stuff.h"
#include "queue.h"
/*******************************************************************************/
MPID_Protocol *MPID_CH_Self_setup(void);
/*-----------------------------------------------------------------------------*/
int MPID_CH_Self_send( void *buf, 
		       int len, 
		       int src_lrank, 
		       int tag, 
		       int context_id, 
		       int dest, 
		       MPID_Msgrep_t msgrep,
		       MPIR_SHANDLE *shandle,
		       p_mad_channel_t channel );
/*------------------------------------------------------------------------------*/
int MPID_CH_Self_isend( void *buf,  
			int len,
			int src_lrank, 
			int tag, 
			int context_id, 
			int dest, 
			MPID_Msgrep_t msgrep,
			MPIR_SHANDLE *shandle,
			p_mad_channel_t channel );
/*------------------------------------------------------------------------------*/
void MPID_CH_Self_delete( MPID_Protocol * );
/*-------------------------------------------------------------------------------*/
int MPID_CH_Self_Check_incoming( MPID_Device *dev, 
				 MPID_BLOCKING_TYPE blocking );
/*-------------------------------------------------------------------------------*/
int MPID_CH_Self_Abort( struct MPIR_COMMUNICATOR *comm_ptr, 
			int code, 
			char *msg );
/*-------------------------------------------------------------------------------*/
int MPID_CH_Self_End( MPID_Device * dev );
/*-------------------------------------------------------------------------------*/
int MPID_CH_Self_push( MPIR_RHANDLE *rhandle, 
		       void *in_runex );
/*-------------------------------------------------------------------------------*/

MPID_Device *MPID_CH_InitSelfMsg( int * argc,
				  char ***argv, 
				  int short_len, 
				  int long_len )
{
  MPID_Device *dev = NEW( MPID_Device );

  if (!dev) 
    {
      return NULL;
    }  
  
  dev->long_len      = 0;
  dev->vlong_len     = 0;
  dev->channel_table = NULL;
  dev->short_msg     = MPID_CH_Self_setup();
  dev->long_msg      = dev->short_msg;
  dev->vlong_msg     = dev->short_msg;
  dev->eager         = dev->short_msg;
  dev->rndv          = dev->short_msg;
  dev->check_device  = NULL; 
  dev->terminate     = MPID_CH_Self_End;
  dev->abort	     = MPID_CH_Self_Abort;
  dev->next	     = NULL;
  
  /* Set the file for Debugging output.  The actual output is controlled
     by MPIDDebugFlag */
  return dev;
}
/*---------------------------------------------------------------------------------*/
MPID_Protocol *MPID_CH_Self_setup(void)
{
    MPID_Protocol *p = NEW( MPID_Protocol );

    if (!p) 
      {
	return NULL;
      }

    p->send	   = MPID_CH_Self_send;
    p->recv	   = NULL;
    p->isend	   = MPID_CH_Self_isend;
    p->wait_send   = NULL;
    p->push_send   = NULL;
    p->cancel_send = NULL;
    p->irecv	   = NULL;
    p->wait_recv   = NULL;
    p->push_recv   = NULL;
    p->cancel_recv = NULL;
    p->do_ack      = NULL;
    p->unex        = NULL; /* could be save, but directly called from send */
    p->delete      = MPID_CH_Self_delete;

    return p;
}
/*-----------------------------------------------------------------------------------*/
int MPID_CH_Self_abort( MPIR_SHANDLE *shandle )
{
  marcel_fprintf(stderr,"Error: blocking send or recv on a self communication\n");
  exit(1);
}
/*-----------------------------------------------------------------------------------*/
int MPID_CH_Self_isend( void *buf, 
			int len, 
			int src_lrank, 
			int tag, 
			int context_id, 
			int dest, 
			MPID_Msgrep_t msgrep,
			MPIR_SHANDLE *shandle,
			p_mad_channel_t channel )
{
  MPIR_RHANDLE *rhandle;
  int          is_posted;
  
  MPID_Msg_arrived( src_lrank,
		    tag, 
		    context_id,
		    &rhandle, 
		    &is_posted ); 
  
  if (is_posted) 
    {
      int msglen = len;
      int err    = 0;

      /* FIXME: what should be done there in case rhandle->len < msglen */
      MPID_HANDLE_MUTEX_LOCK( rhandle );
      MPID_CHK_MSGLEN(rhandle,msglen,err);
      MEMCPY(rhandle->buf,buf,msglen);  
      rhandle->s.MPI_TAG    = tag;
      rhandle->s.MPI_SOURCE = src_lrank;
      rhandle->s.MPI_ERROR  = err;
      rhandle->s.count      = msglen;
      MPID_HANDLE_STATUS_READY( rhandle );   /* syncro point */
      if(rhandle->finish)
	{
	  (*(rhandle->finish))( rhandle );
	}
      MPID_HANDLE_MUTEX_UNLOCK( rhandle );
      MPID_HANDLE_COMPLETE_READY( rhandle ); /* syncro point */

      MPID_HANDLE_MUTEX_LOCK( shandle );
      if(shandle->finish)
	{
	  (*(shandle->finish))( shandle );
	}
      MPID_HANDLE_MUTEX_UNLOCK( shandle );
      MPID_HANDLE_COMPLETE_READY( shandle ); /* syncro point */

      BROADCAST_LOOP(); 
    } 
  else 
    {
      MPID_HANDLE_MUTEX_LOCK( shandle );
      shandle->wait            = MPID_CH_Self_abort;
      shandle->test            = NULL;
      MPID_HANDLE_COMPLETE_RESET( shandle );  
      shandle->start           = buf;
      shandle->bytes_as_contig = len;
      MPID_HANDLE_MUTEX_UNLOCK( shandle );
      MPID_HANDLE_MUTEX_LOCK( rhandle );
      rhandle->s.MPI_TAG       = tag;
      rhandle->s.MPI_SOURCE    = src_lrank;
      rhandle->s.MPI_ERROR     = 0; /* MPI_SUCCESS */
      rhandle->s.count         = len;
      MPID_HANDLE_STATUS_READY( rhandle ); /* syncro point */
      rhandle->from            = MPID_MyWorldRank; /* Needed for flow control */
      MPID_HANDLE_COMPLETE_RESET( rhandle );  
      rhandle->send_id         = shandle;
      rhandle->push            = MPID_CH_Self_push;
      MPID_HANDLE_PUSH_READY( rhandle );        /* syncro point */
      MPID_HANDLE_MUTEX_UNLOCK( rhandle );
    }

  return 0;
}
/*---------------------------------------------------------------------------*/
int MPID_CH_Self_send( void *buf, 
		       int len, 
		       int src_lrank, 
		       int tag, 
		       int context_id, 
		       int dest, 
		       MPID_Msgrep_t msgrep, 
		       MPIR_SHANDLE *shandle,
		       p_mad_channel_t channel )
{
  MPIR_RHANDLE *rhandle;
  int is_posted;
  
  MPID_Msg_arrived( src_lrank, 
		    tag, 
		    context_id,
		    &rhandle, 
		    &is_posted ); 
  
  if (is_posted) 
    {
      int msglen = len;
      int err    = 0;
      
      MPID_HANDLE_MUTEX_LOCK( rhandle );
      /* FIXME: what should be do there in case rhandle->len < msglen */
      MPID_CHK_MSGLEN(rhandle,msglen,err);
      MEMCPY(rhandle->buf,buf,msglen); 
      rhandle->s.MPI_TAG    = tag;
      rhandle->s.MPI_SOURCE = src_lrank;
      rhandle->s.MPI_ERROR  = err;
      rhandle->s.count      = msglen;
      MPID_HANDLE_STATUS_READY( rhandle );   /* syncro point */
      if(rhandle->finish)
	{
	  (*(rhandle->finish))( rhandle );
	}
      MPID_HANDLE_MUTEX_UNLOCK( rhandle );
      MPID_HANDLE_COMPLETE_READY( rhandle ); /* syncro point */
      BROADCAST_LOOP();
    } 
  else
    { 
      MPID_CH_Self_abort(NULL);  
    }
  return MPI_SUCCESS; 
}
/*-------------------------------------------------------------------------*/
int MPID_CH_Self_push( MPIR_RHANDLE *rhandle, 
		       void *in_runex )
{
  MPIR_RHANDLE *runex      = (MPIR_RHANDLE *)in_runex;
  MPIR_SHANDLE *shandle    = (MPIR_SHANDLE *)runex->send_id;
  int          msglen, err = 0;
  
  MPID_HANDLE_MUTEX_LOCK( runex );
  MPID_HANDLE_MUTEX_LOCK( rhandle );
  MPID_HANDLE_MUTEX_LOCK( shandle );
  msglen = runex->s.count;
  MPID_CHK_MSGLEN(rhandle,msglen,err);
  
  /* Copy the data from the local area and free that area */
  if (runex->s.count > 0) 
    {
      MEMCPY( rhandle->buf, shandle->start, msglen );
    }
  MPID_HANDLE_COMPLETE_READY( shandle ); /* syncro point */
  MPID_HANDLE_MUTEX_UNLOCK( shandle );
  
  /* MPID_FLOW_MEM_RECV(msglen,runex->from); */
  
  MPID_DO_HETERO(rhandle->msgrep = runex->msgrep);
  rhandle->s		 = runex->s;
  MPID_HANDLE_MUTEX_UNLOCK( runex ); 
  rhandle->s.MPI_ERROR = err;
  rhandle->wait	 = NULL;
  rhandle->test	 = NULL;
  rhandle->push	 = NULL;
  if(rhandle->finish)
    {
      (*(rhandle->finish))( rhandle );
    }
  MPID_HANDLE_MUTEX_UNLOCK( rhandle );
  MPID_HANDLE_COMPLETE_READY( rhandle ); /* syncro point */
  BROADCAST_LOOP();

  MPID_RecvFree( runex );
  return err;
}
/*-----------------------------------------------------------------------------*/
void MPID_CH_Self_delete( MPID_Protocol *p )
{
  FREE( p );
}
/*-----------------------------------------------------------------------------*/
int MPID_CH_Self_End( MPID_Device *dev )
{
  DEBUG_PRINT_MSG("Entering MPID_CH_Self_end\n");
  /* Finish off any pending transactions */
  
  if (MPID_GetMsgDebugFlag()) {
    MPID_PrintMsgDebug();
  }
  
  (*(dev->short_msg->delete))( dev->short_msg );
  FREE( dev );
  
  /* We should really generate an error or warning message if there 
     are uncompleted operations... */
  
  return 0;
}
/*------------------------------------------------------------------------------*/
int MPID_CH_Self_Abort(
		       struct MPIR_COMMUNICATOR *comm_ptr,
		       int      code,
		       char     *msg
		       )
{
  if (msg)
    {
      marcel_fprintf( stderr, "[%d] %s\n", MPID_MyWorldRank, msg );
    }
  else
    {
      marcel_fprintf( stderr, "[%d] Aborting program!\n", MPID_MyWorldRank );
    }
  marcel_fflush( stderr );
  marcel_fflush( stdout );
  
  exit(0);  
  return 0;
}









