
#ifndef _GENERIC_H
#define _GENERIC_H

#include "mpid.h"
#include "mpiddev.h"
#include "mpimem.h"
#include "syncro.h"

/*********************************************************************************/
/*  Those structures are used for function calls taking more than 1 argument     */
/*  in the marcel_create function.                                               */
/*********************************************************************************/

typedef union{
  MPID_PKT_OK_TO_SEND_T   sendok_pkt;
  MPID_PKT_ANTI_SEND_T    antisend_pkt;
} CH_MAD_PKT_t;

typedef union{
  MPID_Aint               rhandle_address; 
  MPID_PKT_HEAD_T         head_pkt;
  MPID_PKT_REQUEST_SEND_T request_pkt;
  MPID_PKT_OK_TO_SEND_T   sendok_pkt;
  MPID_PKT_ANTI_SEND_T    antisend_pkt;
} CH_MAD_HEAD_FIELD_t;

typedef enum {
  MAD_SHORT_PKT     = 111,
  MAD_REQUEST_PKT   = 222,
  MAD_SENDOK_PKT    = 333, 
  MAD_CANCEL_PKT    = 444,
  MAD_RNDV_PKT      = 666,
  MAD_ABORT_PKT     = 777,
  MAD_TERM_PKT      = 888
} CH_MAD_MESSAGE_TYPE;

/*---------------------------------------------------------------------*/

typedef struct{               /* See the  MPID_CH_Mad_Rndv_isend function        */
  void            *buf;       /* in madrndv.c                                    */ 
  int             len;        /* See the  MPID_CH_Mad_eager_isend function       */ 
  int             src_lrank;  /* in madshort.c                                   */
  int             tag;
  int             context_id;
  int             dest;
  MPID_Msgrep_t   msgrep;
  MPIR_SHANDLE    *shandle;
  p_mad_channel_t channel;
} thread_args;

typedef struct{              /* See the MPID_CH_Mad_Rndv_ok_to_send function    */
  MPID_Aint        send_id;  /* in madrndv.c                             */  
  MPID_Aint        recv_id;
  int              from;  
  p_mad_channel_t  channel;
} thread_args2;

typedef struct{             /* See the MPID_CH_Mad_Rndv_ack  function          */
  CH_MAD_PKT_t    in_pkt; /* in madrndv.c                        */  
  int             from_grank;
  p_mad_channel_t channel;
} thread_args3;

typedef struct{                     /* See the MPID_CH_Mad_Poll  function              */
  MPID_Device       *device;        /* in mpimad.c                    */ 
  p_mad_channel_t   channel;
  int               channel_number;
  MPID_PKT_T        pkt; 
  /* p_tbx_memory_t fast_buffer; */
} poll_args;

// this size should be kept low, ie up to 16 bytes on 32-bits
// systems  
#define OPT
#ifdef OPT
#define OPT_BUF_SIZE 8
#endif

typedef struct {
   CH_MAD_MESSAGE_TYPE  type;
   CH_MAD_HEAD_FIELD_t  opt_field;
#ifdef OPT
   char                 padding[OPT_BUF_SIZE];
#endif
#if defined(GLOBAL_DEBUG) || defined(DEBUG_POLLING)
   int                  node_id;
   int                  tid;    
#endif   
} CH_MAD_HEADER_t; 


typedef struct {
   MPID_PKT_SHORT_T   *pkt;
   p_mad_connection_t  connection;
   void               *buffer;
  /*p_tbx_memory_t     fast_buffer;*/
} CH_MAD_SHORT_ARGS;

#define NO_CONNECTION NULL
#endif /* GENERIC_H*/














