
/* benchmarking/switch value measurement
   this has to be defined before including packets.h where MPID_PKT_MAX_DATA_SIZE is defined
   #define BENCH */

#include "generic.h" 
#include "reqalloc.h"
#include <stdio.h>

#define ENFORCE_ALIGNMENT  
#define MPID_RNDV_SELF

#ifdef VERBOSE
#define OUTPUT(string) marcel_fprintf(stderr,"[Node %i]:\t%s\n",MPID_MyWorldRank,string)
#define OUTPUTF(string, args...) if(1){ \
	char buffer[256]; \
	marcel_snprintf(buffer, 256, "[Node %i]:\t%s\n",MPID_MyWorldRank,string); \
	marcel_fprintf(stderr, buffer, ## args); \
	}
#else
#define OUTPUT(string)
#define OUTPUTF(s, args...)
#endif

extern CH_MAD_channel_info_t *MPID_MAD_Global_Channel_Table;
extern CH_MAD_channel_info_t *MPID_MAD_Channel_Table;
extern p_mad_channel_t       *MPID_MAD_Default_Channel;
extern MPI_Comm              *MPI_USER_COMM; 
extern int                    MPI_COMM_WORLD_INDEX;

/* 
   those sizes MUST be compatible with the 
   MPID_PKT_MAX_DATA_SIZE constant defined in packets.h ( 128*1024 for now ) 
   */

#ifdef MARCEL_SMP
#define TCP_SWITCH                128*1024 //300
#define SCI_SWITCH                122*1024        
#else  /* MARCEL_SMP */
#define TCP_SWITCH                128*1024       
#define SCI_SWITCH                 32*1024
#endif /* MARCEL_SMP */

#define BIP_SWITCH                 65*1024       
#define HETERO_SWITCH             128*1024       
#define DEFAULT_SWITCH            128*1024  
#define DEFAULT_TYPE_SWITCH DEFAULT_SWITCH 

// Unused for now : the tradeoff should be made between 
// memory use vs speed. 
#define GLOBAL_SPACE_SIZE      2*1024*1024 /* total size for buffering */
#define BUFFER_NUMBER          ((GLOBAL_SPACE_SIZE) / (MPID_PKT_MAX_DATA_SIZE)) 
                                           /* number of buffers        */


#ifdef  ENFORCE_ALIGNMENT
#define SIZE_OF_MADHEADER      tbx_aligned(sizeof(CH_MAD_HEADER_t), MAD_LENGTH_ALIGNMENT)
#else   
#define SIZE_OF_MADHEADER      sizeof(CH_MAD_HEADER_t)
#define TYPE_SWITCH            DEFAULT_SWITCH
#endif


/*===================================================================*/
/* those are seen from the ADI */
/*-------------------------------------------------------------------*/
void MPID_CH_Mad_SendCancelPkt( MPI_Request *, 
				int *);
/*-------------------------------------------------------------------*/
int MPID_CH_Mad_Rndv_isend( void *,
			    int,
			    int,
			    int,
			    int,
			    int,
			    MPID_Msgrep_t,
			    MPIR_SHANDLE *,
			    p_mad_channel_t );
/*-------------------------------------------------------------------*/
int MPID_CH_Mad_Rndv_sendtype( void *,
			       int,
			       struct MPIR_DATATYPE *,
			       int,
			       int,
			       int,
			       int,
			       MPID_Msgrep_t,
			       p_mad_channel_t );

/*-----------------------------------------------------------------------------*/
int MPID_CommInit( struct MPIR_COMMUNICATOR *, struct MPIR_COMMUNICATOR *);
/*-----------------------------------------------------------------------------*/
int MPID_Comm_Override_Channel( struct MPIR_COMMUNICATOR *, char *);
/*-----------------------------------------------------------------------------*/
marcel_mutex_t* Mutex_From_Channel( p_mad_channel_t, CH_MAD_channel_info_t *);
/*-----------------------------------------------------------------------------*/
int Threshold_From_Channel( p_mad_channel_t , CH_MAD_channel_info_t *);
/*-----------------------------------------------------------------------------*/
void MPID_CH_Mad_Init_Channel_info( CH_MAD_channel_info_t *channel_list, 
				    int length );
/*-----------------------------------------------------------------------------*/
p_mad_channel_t MPID_CH_Mad_Default_Channel(CH_MAD_channel_info_t *channel_list);
/*-----------------------------------------------------------------------------*/
p_mad_channel_t MPID_CH_Mad_Get_Default_Channel( CH_MAD_channel_info_t *);
/*-----------------------------------------------------------------------------*/
ntbx_process_lrank_t MPID_CH_Mad_Generate_Key( p_mad_channel_t );
/*-----------------------------------------------------------------------------*/
ntbx_process_lrank_t MPID_CH_Mad_Global_to_Local( p_mad_channel_t , int );
/*-----------------------------------------------------------------------------*/
void MPID_Generate_Communicators( void );
/*-----------------------------------------------------------------------------*/
void MPID_Set_lranks ( struct MPIR_COMMUNICATOR * );
/*-----------------------------------------------------------------------------*/
#define MASTER_OF_CHANNEL( channel ) (MPID_CH_Mad_Generate_Key( channel )== 0) ? 1 : 0
#define MPID_MAD_CHANNEL_SIZE( channel ) ( channel )?(( channel )->pc->count) : 0
#define MPID_Get_Channel_From_Communicator( comm_ptr ) (( comm_ptr ) && (( comm_ptr )->channel )) \
                                                      ?(( comm_ptr )->channel) : NULL 



















