
/***********************************************************************************/
#include "hot_stuff.h"
/***********************************************************************************/
/***********************************************************************************/
/*                                   PROTOTYPES                                    */
/***********************************************************************************/
/*---------------------------------------------------------------------------------*/
p_mad_channel_t MPID_CH_Mad_Get_Channel_by_name( char *, 
		                              	 CH_MAD_channel_info_t *,
						 int );
/*---------------------------------------------------------------------------------*/
p_mad_channel_t MPID_CH_Mad_Default_Channel( CH_MAD_channel_info_t * );
/*---------------------------------------------------------------------------------*/
p_mad_channel_t MPID_CH_Mad_Get_Default_Channel( CH_MAD_channel_info_t * );
/*---------------------------------------------------------------------------------*/
marcel_mutex_t* Mutex_From_Channel( p_mad_channel_t,
				    CH_MAD_channel_info_t * );
/*---------------------------------------------------------------------------------*/
int  Threshold_From_Channel( p_mad_channel_t,
			     CH_MAD_channel_info_t *);
/*---------------------------------------------------------------------------------*/
void MPID_CH_Mad_Init_Channel_info( CH_MAD_channel_info_t *, int );
/*---------------------------------------------------------------------------------*/
int MPID_CommInit( struct MPIR_COMMUNICATOR *, 
		   struct MPIR_COMMUNICATOR * );
/*---------------------------------------------------------------------------------*/
int MPID_CommFree( struct MPIR_COMMUNICATOR * );
/*---------------------------------------------------------------------------------*/
int MPID_Comm_Override_Channel( struct MPIR_COMMUNICATOR *, 
				char *);
/*---------------------------------------------------------------------------------*/
ntbx_process_lrank_t MPID_CH_Mad_Generate_Key( p_mad_channel_t );
/*---------------------------------------------------------------------------------*/
ntbx_process_lrank_t MPID_CH_Mad_Global_to_Local( p_mad_channel_t , int );
/*---------------------------------------------------------------------------------*/
void MPID_Generate_Communicators( void );
/*---------------------------------------------------------------------------------*/
void MPID_Set_lranks ( struct MPIR_COMMUNICATOR *);
/*---------------------------------------------------------------------------------*/
/***********************************************************************************/
/*                                   DEFINITIONS                                   */
/***********************************************************************************/
/*---------------------------------------------------------------------------------*/
p_mad_channel_t MPID_CH_Mad_Get_Channel_by_name( char                  *channel_id , 
		                              	 CH_MAD_channel_info_t *channel_list , 
						 int                    length)
{
  int i = 0;
  
  while ( i < length  )
    {
      if (strcmp( (channel_list[i]).channel_id, channel_id ) == 0) 
	{
	  return (channel_list[i]).channel;
	}
      else
	{
	  i++;
	}
    }
  return NULL;
}
/*-----------------------------------------------------------------------------------*/
/* MPID_CHANNELS_COLORS MPID_CH_Mad_Channel_Member (char                  *channel_id ,  */
/* 						 CH_MAD_channel_info_t *channel_list)  */
/* { */
/*   int i = 0; */
/*   while ( i < MPID_MAD_Channel_Number  ) */
/*     { */
/*       if (strcmp( (channel_list[i]).channel_id, channel_id ) == 0)  */
/* 	{ */
/* 	  return (channel_list[i]).channel_color; */
/* 	} */
/*       else */
/* 	{ */
/* 	  i++; */
/* 	} */
/*     } */
/*   return MPID_NO_COLOR; */
/* } */
/*-----------------------------------------------------------------------------------*/
p_mad_channel_t MPID_CH_Mad_Default_Channel( CH_MAD_channel_info_t *channel_list )
{
   int                   index , index2;
   int                   candidats = 0;
   CH_MAD_channel_info_t temporary[MPID_MAD_Channel_Number];
   
   for( index = 0, index2 =0 ; index < MPID_MAD_Channel_Number; index ++)
     {
	if((MPID_MAD_CHANNEL_SIZE(channel_list[index].channel)) == MPID_MyWorldSize )
	  {
	     temporary[index2++] = channel_list[index];
	     candidats++;
	  }
     }
   if (candidats)
     {
	for( index = 0 ; index < candidats; index ++)
	  {
	     p_mad_channel_t channel_address = 
	       MPID_CH_Mad_Get_Channel_by_name( "default", temporary, candidats );
	     if ( channel_address )
	       {
		  return channel_address;
	       }
	     else
	       {
		  channel_address = MPID_CH_Mad_Get_Channel_by_name( "tcp", temporary, candidats );
		  if (channel_address)
		    {
		       return channel_address;
		    }
		  else
		    {
		       return temporary[0].channel;
		    }
	       }
	  }
     }
   else
     {
	marcel_fprintf(stderr,"MPICH/MAD3 ERROR : No default channel found \n");
	marcel_fflush(stderr);
	abort();
	return NULL;
     }
   // never reached, avoids compiler warning
   return NULL;
}
/*----------------------------------------------------------------------------------*/
p_mad_channel_t MPID_CH_Mad_Get_Default_Channel( CH_MAD_channel_info_t *channel_list )
{
  return *MPID_MAD_Default_Channel;
}
/*-----------------------------------------------------------------------------------*/
marcel_mutex_t* Mutex_From_Channel( p_mad_channel_t        channel,
				    CH_MAD_channel_info_t *channel_list )
{
  int i = 0;
  
  while ( i < MPID_MAD_Channel_Number )
    {
      if ((channel_list[i].channel) == channel) 
	{
	  return &((channel_list[i]).send_ok);
	}
      else
	{
	  i++;
	}
    }
#ifdef GLOBAL_DEBUG
  marcel_printf("Channel not found !!!\n"); 
#endif
  return NULL;
}
/*-----------------------------------------------------------------------------------*/
int Threshold_From_Channel( p_mad_channel_t        channel,
			    CH_MAD_channel_info_t *channel_list )
{
   int i = 0;
  
   while ( i < MPID_MAD_Channel_Number )
     {
	if ((channel_list[i].channel) == channel) 
	  {
	     return ((channel_list[i]).channel_threshold);
	  }
	else
	  {
	     i++;
	  }
     }
#ifdef GLOBAL_DEBUG
   marcel_printf(">>>>> Channel not found  ! <<<<<\n");
#endif  
   return -1;
}
/*-----------------------------------------------------------------------------------*/
void MPID_CH_Mad_Init_Channel_info( CH_MAD_channel_info_t *channel_list , int length)
{
  int i;

  for (i = 0 ; i < length ; i++)
    {
      if ( strcmp( (channel_list[i]).channel_id , "default" ) == 0)
	{
	  (channel_list[i]).channel_threshold = DEFAULT_SWITCH;
	  (channel_list[i]).channel_color     = MPID_DEFAULT;
	}
      else if ( strcmp( (channel_list[i]).channel_id , "tcp" ) == 0)
	{
	  (channel_list[i]).channel_threshold = TCP_SWITCH; 
	  (channel_list[i]).channel_color     = MPID_TCP;
	}
      else if ( strcmp( (channel_list[i]).channel_id , "sci" ) == 0)
	{
	  (channel_list[i]).channel_threshold = SCI_SWITCH;  
	  (channel_list[i]).channel_color     = MPID_SCI;
	}
      else if ( strcmp( (channel_list[i]).channel_id , "bip" ) == 0)
	{
	  (channel_list[i]).channel_threshold = BIP_SWITCH;
	  (channel_list[i]).channel_color     = MPID_BIP; 
	}
      else if ( strcmp( (channel_list[i]).channel_id , "hetero" ) == 0)
	{
	  (channel_list[i]).channel_threshold = HETERO_SWITCH;
	  (channel_list[i]).channel_color     = MPID_HETERO; 
	}
      else
	{
	  (channel_list[i]).channel_threshold = DEFAULT_SWITCH;
	  (channel_list[i]).channel_color     = MPID_DEFAULT; 
	}
    }
}
/*---------------------------------------------------------------------------------*/
int MPID_CommInit( struct MPIR_COMMUNICATOR *comm1, 
		   struct MPIR_COMMUNICATOR *comm2 )
{
  if(!comm2)
    {
      return MPI_SUCCESS;
    }
  else
    {
      MPID_THREAD_DS_LOCK_INIT( comm2 ); 
      if ( comm1 )
	{
	  comm2->channel = comm1->channel; 
	}
      else
	{
	  comm2->channel = MPID_CH_Mad_Get_Default_Channel( MPID_MAD_Channel_Table );
	}
      
      if ( !comm2->channel )
	{
	  return 1;
	}
      else
	{
	  return MPI_SUCCESS;
	}
    }
}
/*----------------------------------------------------------------------------------*/
int MPID_CommFree( struct MPIR_COMMUNICATOR *comm )
{
  /* destroy the mutex I guess ... */
  return MPI_SUCCESS;
}
/*----------------------------------------------------------------------------------*/
int MPID_Comm_Override_Channel( struct MPIR_COMMUNICATOR *comm , char *channel_id )
{
  p_mad_channel_t channel = MPID_CH_Mad_Get_Channel_by_name( channel_id,
							     MPID_MAD_Channel_Table,
							     MPID_MAD_Channel_Number);
  if (comm)
    { 
      if ( channel ) 
	{
	  comm->channel = channel;
	  return 0;
	}
      else
	{
	  comm->channel = MPID_CH_Mad_Get_Default_Channel( MPID_MAD_Channel_Table ); 
#ifdef GLOBAL_DEBUG
	  marcel_printf(">>>>> Channel Override in Communicator failed,
                               switching to default channel <<<<<\n");
#endif
	  return 1;
	}
    } 
  else
    {
#ifdef GLOBAL_DEBUG
      marcel_printf(">>>>> Channel Override in NULL Communicator not allowed ! <<<<<\n");
#endif
      return 1;
    }
}
/*----------------------------------------------------------------------------------*/
ntbx_process_lrank_t MPID_CH_Mad_Generate_Key( p_mad_channel_t channel )
{
  p_ntbx_process_container_t pc            = channel->pc;
  ntbx_process_lrank_t       process_lrank = ntbx_pc_global_to_local( pc , MPID_MyWorldRank );
  
  return process_lrank;
}
/*----------------------------------------------------------------------------------*/
ntbx_process_lrank_t MPID_CH_Mad_Global_to_Local( p_mad_channel_t channel, int grank )
{
  p_ntbx_process_container_t pc            = channel->pc;
  ntbx_process_lrank_t       process_lrank = ntbx_pc_global_to_local( pc , grank );
  
  return process_lrank;
}
/*----------------------------------------------------------------------------------*/
ntbx_process_grank_t MPID_CH_Mad_Reverse_Key( p_mad_channel_t channel ,int lrank)
{
  p_ntbx_process_container_t pc            = channel->pc;
  ntbx_process_grank_t       process_grank = ntbx_pc_local_to_global( pc , lrank );
  
  return process_grank;
}
/*----------------------------------------------------------------------------------*/
void MPID_Generate_Communicators( void )
{
  p_mad_channel_t default_channel = 
    MPID_CH_Mad_Get_Default_Channel( MPID_MAD_Global_Channel_Table );
  int             index, index2 = 0;
  int             color;
  int             key;
  MPI_Comm        comm;

  if( MPID_MAD_Global_Channel_Number == 1)
    {
      MPI_USER_COMM[0] = MPI_COMM_WORLD;
    }
  else
    {
      marcel_fprintf(stdout,"Generating communicators ...");
      
      for ( index= 0; index < MPID_MAD_Global_Channel_Number; index++ )
	{
	  if((MPID_MAD_Global_Channel_Table[index]).channel != default_channel) 
	    { 	  
	      if((MPID_MAD_Global_Channel_Table[index]).channel == NULL)
		{
		  color = MPID_NO_COLOR;
		  key   = MPID_MyWorldRank;   
		}
	      else
		{
		  color = (MPID_MAD_Global_Channel_Table[index]).channel_color;
		  key   = MPID_CH_Mad_Generate_Key((MPID_MAD_Global_Channel_Table[index]).channel);
		}
	      
	      MPI_Comm_split(MPI_COMM_WORLD , color, key , &comm);
	      
	      if ( comm != MPI_COMM_NULL)
		{
		  MPI_USER_COMM[index2++] = comm;
		}
	    }
	  else
	    {
	      MPI_USER_COMM[index2] = MPI_COMM_WORLD;
	      MPI_COMM_WORLD_INDEX  = index2++;
	    }
	}
       
       
       for ( index = 0 ; index < MPID_MAD_Channel_Number; index++)
	 {
	    if((MPID_MAD_Channel_Table[index]).channel != default_channel)
	     {
		struct MPIR_COMMUNICATOR *comm_ptr = MPIR_GET_COMM_PTR( MPI_USER_COMM[index] );
		comm_ptr->channel                  = (MPID_MAD_Channel_Table[index]).channel ;
		MPID_Set_lranks ( comm_ptr ); 
	     }
	 }
       marcel_printf("Done\n");    
    }
}
/*----------------------------------------------------------------------------------*/
/* should this routine be called in MPID_CommInit ?? */
void MPID_Set_lranks ( struct MPIR_COMMUNICATOR *comm_ptr )
{
   struct MPIR_GROUP *group  = comm_ptr->group ;
   int               nb_proc = group->np;
   int               index;
   
   // Alright, it's a bit tricky here: I have to make a double 
   // conversion for the indexes. The first happens at the group/Communicator
   // creation (ie now). The reverse conversion step is involved just before 
   // sending anything. See any sending primitive to catch this.
   for (index = 0; index < nb_proc ; index++) 
     {
	group->lrank_to_grank[index] =  MPID_CH_Mad_Reverse_Key( comm_ptr->channel, 
								 index);
     }
   
   group->local_rank =  MPID_CH_Mad_Generate_Key( comm_ptr->channel ); 
   
   if (group->local_rank >= nb_proc)
     {
	group->local_rank = MPI_UNDEFINED;
     }
}
/*----------------------------------------------------------------------------------*/














