#include <assert.h>
#include "smpi.h"
#include "shell.c"
#include "mpiddev.h"
#include "mpimem.h"

extern int MPID_MyWorldSize;
extern int MPID_MyWorldRank;
extern MPID_Device *MPID_CH_InitSelfMsg 
       ANSI_ARGS(( int *, char ***, int, int ));

extern p_mad_madeleine_t madeleine;
#define OUTPUT(string) fprintf(stderr,"[Node %i]:\t%s\n",MPID_MyWorldRank,string) 

p_mad_dir_node_t find_my_node()
{
  p_tbx_slist_t list;
  p_mad_dir_node_t tnode;
  ntbx_process_grank_t res;
  int found = 0;

  list = madeleine->dir->node_slist;

  tbx_slist_ref_to_head(list);
  
  do
  {

	tnode = tbx_slist_ref_get(list);

	ntbx_pc_first_global_rank(tnode->pc, &res); // on imagine que ce n'est pas vide
	if(res == MPID_MyWorldRank)
		found = 1;
	else
	{
		while(ntbx_pc_next_global_rank(tnode->pc, &res))
		{
			if(res == MPID_MyWorldRank)
			{
				found = 1;
				break;
			}
		}
	}
  } while(!found && tbx_slist_ref_forward(list));

  return tnode;
}

MPID_Config * MPID_GetConfigInfo(int  *argc, char ***argv, void *config,int  *error_code)
{
  MPID_Config *new_config = NULL;
  MPID_Config *return_config = NULL;
  int i, j;
  p_mad_dir_node_t tnode;
  ntbx_process_grank_t res;

  if (config != NULL)
    return (MPID_Config *)config;

  /* search the number of local processes */

  tnode = find_my_node();

  OUTPUT("Physical Node found");

  smpi.num_local_nodes = 0;
  
  for (i = 0; i < MPID_MyWorldSize; i++)
    smpi.local_nodes[i] = -1;

  ntbx_pc_first_global_rank(tnode->pc, &res);
 
  do
  {
  	if(res == MPID_MyWorldRank)
	  smpi.my_local_id = smpi.num_local_nodes;
	smpi.local_nodes[res] = smpi.num_local_nodes;
	smpi.num_local_nodes++;
  } while(ntbx_pc_next_global_rank(tnode->pc, &res));
   
  assert(smpi.num_local_nodes != 0);  

  /* Madeleine II  device */
  if (MPID_MyWorldSize > smpi.num_local_nodes) {
    if (new_config == NULL) {
      new_config = (MPID_Config *)MALLOC( sizeof(MPID_Config) );
      assert(new_config);
      return_config = new_config;
    } else {
      new_config->next = (MPID_Config *)MALLOC( sizeof(MPID_Config) );
      assert(new_config->next);
      new_config = new_config->next;
    }
    new_config->device_init = MPID_CH_InitMsgPass;
    new_config->device_init_name = (char *)malloc(255 * sizeof(char));
    sprintf(new_config->device_init_name, "Madeleine II device");
    new_config->num_served = MPID_MyWorldSize - smpi.num_local_nodes;
    new_config->granks_served = (int *)malloc(new_config->num_served * sizeof(int));
    j=0;
    for(i=0; i<MPID_MyWorldSize; i++)
      if (smpi.local_nodes[i] == -1) {
	new_config->granks_served[j] = i;
	j++;
      }
    assert(j == new_config->num_served);
    new_config->next = NULL;
  }
  else {
    fprintf(stderr,"Local processes only...\n");
//    mad_init(argc,*argv);
//    fprintf(stderr,"Mad Init done ...\n");
  }

  /* Self device */
  if (new_config == NULL) {
    new_config = (MPID_Config *)MALLOC( sizeof(MPID_Config) );
    assert(new_config);
    return_config = new_config;
  } else {
    new_config->next = (MPID_Config *)MALLOC( sizeof(MPID_Config) );
    assert(new_config->next);
    new_config = new_config->next;
  }
  new_config->device_init = MPID_CH_InitSelfMsg;
  new_config->device_init_name = (char *)malloc(255 * sizeof(char));
  sprintf(new_config->device_init_name, "Self device");
  new_config->num_served = 1;
  new_config->granks_served = (int *)malloc(sizeof(int));
  new_config->granks_served[0] = MPID_MyWorldRank;
  new_config->next = NULL;


  /* Smp device */
  if (smpi.num_local_nodes > 1) {
    if (new_config == NULL) {
      new_config = (MPID_Config *)MALLOC( sizeof(MPID_Config) );
      assert(new_config);
      return_config = new_config;
    } else {
      new_config->next = (MPID_Config *)MALLOC( sizeof(MPID_Config) );
      assert(new_config->next);
      new_config = new_config->next;
    }
    new_config->device_init = MPID_SMP_InitMsgPass;
    new_config->device_init_name = (char *)malloc(255 * sizeof(char));
    sprintf(new_config->device_init_name, "SMP_plug device");
    new_config->num_served = smpi.num_local_nodes - 1;
    new_config->granks_served = (int *)malloc(new_config->num_served * sizeof(int));
    j = 0;
    for(i=0; i<MPID_MyWorldSize; i++)
      if ((i!= MPID_MyWorldRank) && (smpi.local_nodes[i] != -1)) {
	assert(smpi.local_nodes[i] != smpi.my_local_id);
	new_config->granks_served[j] = i;
	j++;
      }
    assert(j == new_config->num_served);
    new_config->next = NULL;
  }

  return return_config;
}








