


/***********************************************************************************/
#include "hot_stuff.h"
#include "queue.h"
/***********************************************************************************/

volatile int expect_cancel_ack = 0;
/***********************************************************************************/
/* 
 * CancelSend 
 * This is fairly hard.  We need to send a "please_cancel_send", 
 * which, if the message is found in the unexpected queue, removes it.
 * However, if the message is being received at the "same" moment, the
 * ok_to_send and cancel_send messages could cross.  To handle this, the
 * receiver must ack the cancel_send message (making the success of the
 * cancel non-local).  There are even more complex protocols, but we won't
 * bother.
 * 
 * Don't forget to update MPID_n_pending as needed.
 */
/***********************************************************************************/
/*---------------------------------------------------------------------------------*/
/* This routine is called from MPI_Cancel.  Its purpose is to send an    
   anti-send packet to the calling process's partner.  If successful,
   the error_code will return MPI_SUCCESS, otherwise the error_code will
   return MPI_ERR_OTHER */

void MPID_CH_Mad_SendCancelPkt( MPI_Request *request, 
				int *err_code )
{  
   
   MPIR_SHANDLE          *shandle = (MPIR_SHANDLE *)*request; 
   CH_MAD_HEADER_t       header   __NET_BUFFER__;
   /*  marcel_mutex_t        *mutex;*/
   p_mad_connection_t    connection;
   p_mad_channel_t       default_channel 
     = MPID_CH_Mad_Default_Channel( MPID_MAD_Channel_Table );
 
   header.type                           = MAD_CANCEL_PKT;   
   header.opt_field.antisend_pkt.mode    = MPID_PKT_ANTI_SEND; 
   header.opt_field.antisend_pkt.lrank   = MPID_MyWorldRank; 
   MPID_HANDLE_MUTEX_LOCK( shandle );
   header.opt_field.antisend_pkt.to      = shandle->partner;
   MPID_HANDLE_MUTEX_UNLOCK( shandle );
   header.opt_field.antisend_pkt.src     = MPID_MyWorldRank;
   MPID_AINT_SET(  header.opt_field.antisend_pkt.send_id , shandle ); 
   
   /* HETERO STUFF */ 
   MPID_PKT_PACK(  header.opt_field.antisend_pkt, 
		   sizeof(  header.opt_field.antisend_pkt ),  
		   header.opt_field.antisend_pkt.to );    
   /*  
    mutex = Mutex_From_Channel( default_channel , MPID_MAD_Channel_Table );
    LOCK_CHANNEL( mutex ); 
    */
   connection = mad_begin_packing( default_channel , header.opt_field.antisend_pkt.to );
   mad_pack( connection, 
	     &header, 
	     SIZE_OF_MADHEADER,
	     mad_send_CHEAPER,
	     mad_receive_EXPRESS);
   mad_end_packing(connection);
   /*  UNLOCK_CHANNEL( mutex ); */
   
   MPID_INCREMENT_CANCEL_ACK();
} 
/*---------------------------------------------------------------------------------*/

/* This routine is called when a process receives an anti-send pkt.  Its 
 purpose is to search for the request found in the pkt in the unexpected
 queue.  If found, set the pkt.cancel to 1, otherwise, set pkt.cancel to 
 0.  Send this information back in an anti-send-ok pkt. */

void MPID_CH_Mad_internal_SendCancelOkPacket( void *in_pkt, int from )
{  

   MPID_PKT_ANTI_SEND_T *pkt        = (MPID_PKT_ANTI_SEND_T *)in_pkt;
   CH_MAD_HEADER_t      header      __NET_BUFFER__;
   /*marcel_mutex_t       *mutex;*/
   p_mad_connection_t   connection;
   p_mad_channel_t      default_channel 
     = MPID_CH_Mad_Default_Channel( MPID_MAD_Channel_Table );
   MPIR_SHANDLE         *shandle    = NULL;
   MPIR_RHANDLE         *rhandle;
   int                  error_code;
   int                  found       = 0; 
   
   /* A cancel packet is a little larger than the basic packet size and 
    may need to be unpacked (in the heterogeneous case) */
   MPID_PKT_UNPACK( (MPID_PKT_HEAD_T *)in_pkt + 1,
		    sizeof(MPID_PKT_ANTI_SEND_T) - sizeof(MPID_PKT_HEAD_T),
		    from ); 
   MPID_AINT_GET(shandle, pkt->send_id); 
   
   /* Look for request, if found, delete it */
   error_code = MPID_Search_unexpected_for_request( shandle, 
						    &rhandle, 
						    &found );
   
   if ( (error_code != MPI_SUCCESS) || (found == 0) ) 
     {
	header.opt_field.antisend_pkt.cancel = 0;
     }
   else 
     {
	int thresh = Threshold_From_Channel( rhandle->channel ,
					     MPID_MAD_Channel_Table );
	MPID_HANDLE_MUTEX_LOCK( rhandle );      
	if ( rhandle->s.count <= thresh ) 
	  {  
	     /* begin if short/eager message */
	     FREE ( rhandle->start ); 
	     rhandle->start = NULL;
	     MPID_HANDLE_MUTEX_UNLOCK( rhandle );
	     MPID_RecvFree( rhandle ); 
	  } 
	else 
	  { /* begin if rndv message */
	     MPID_HANDLE_MUTEX_UNLOCK( rhandle );
	     MPID_RecvFree( rhandle );
	  } 
	header.opt_field.antisend_pkt.cancel = 1;
     }
   
   header.type                           = MAD_CANCEL_PKT;
   header.opt_field.antisend_pkt.mode    = MPID_PKT_ANTI_SEND_OK;
   header.opt_field.antisend_pkt.lrank   = pkt->to;
   header.opt_field.antisend_pkt.to      = from; 
   header.opt_field.antisend_pkt.src     = MPID_MyWorldRank;
   header.opt_field.antisend_pkt.send_id = pkt->send_id;  
   MPID_PKT_PACK( &(header.opt_field.antisend_pkt), 
		  sizeof(header.opt_field.antisend_pkt), from ); 
   
   /*  mutex = Mutex_From_Channel( default_channel , MPID_MAD_Channel_Table );
    LOCK_CHANNEL( mutex ); */
   connection = mad_begin_packing( default_channel , header.opt_field.antisend_pkt.to );
   mad_pack( connection, 
	     &header, 
	     SIZE_OF_MADHEADER,
	     mad_send_CHEAPER,
	     mad_receive_EXPRESS);
   mad_end_packing(connection);
   /*  UNLOCK_CHANNEL( mutex );   */
} 
/*---------------------------------------------------------------------------------*/
void  MPID_CH_Mad_tSendCancelOkPacket( thread_args3 *args )
{
  //marcel_detach(marcel_self());
  
  MPID_CH_Mad_internal_SendCancelOkPacket( &(args->in_pkt), 
					   args->from_grank );
  FREE(args);
}
/*---------------------------------------------------------------------------------*/
void MPID_CH_Mad_SendCancelOkPacket( void *in_pkt, int from )
{
  thread_args3 *args;
  marcel_attr_t attributes; 
   
  args              = NEW( thread_args3 );
  MEMCPY (&(args->in_pkt.antisend_pkt), in_pkt, sizeof(CH_MAD_PKT_t)); 
  args->from_grank  = from;
  args->channel     = NULL;
  
  marcel_attr_init(&attributes);	 
  marcel_attr_setdetachstate(&attributes, TRUE); 
  marcel_create( NULL, &attributes,
		 (marcel_func_t) MPID_CH_Mad_tSendCancelOkPacket, args);  
}
/*---------------------------------------------------------------------------------*/
/* This routine is called when a process receives an anti-send-ok packet.
   If pkt->cancel = 1, then set the request found in the pkt as
   cancelled and complete.  If pkt->cancel = 0, do nothing. */

void MPID_CH_Mad_RecvCancelOkPacket( void *in_pkt, int from )
{  
   MPID_PKT_ANTI_SEND_T *pkt     = (MPID_PKT_ANTI_SEND_T *)in_pkt;
   MPIR_SHANDLE         *shandle = NULL;
  
  /* A cancel packet is a little larger than the basic packet size and 
     may need to be unpacked (in the heterogeneous case) */
  MPID_PKT_UNPACK( (MPID_PKT_HEAD_T *)in_pkt + 1,
		   sizeof(MPID_PKT_ANTI_SEND_T) - sizeof(MPID_PKT_HEAD_T),
		   from );

  MPID_AINT_GET(shandle, pkt->send_id);

  if( pkt->cancel ) 
    {   
      /* Mark the request as cancelled */
      MPID_HANDLE_MUTEX_LOCK( shandle );      
      shandle->s.MPI_TAG = MPIR_MSG_CANCELLED;
      if (shandle->finish)
	{
	  (*(shandle->finish))(shandle);
	}
      shandle->is_cancelled = 1;
      MPID_HANDLE_MUTEX_UNLOCK( shandle );      
      MPID_HANDLE_COMPLETE_READY( shandle );
      MPID_HANDLE_CANCEL_READY( shandle );
      MPID_BREAK_LOOP();
      MPID_DECREMENT_PENDING();  
      DEBUG_PRINT_MSG("Request has been successfully cancelled");
    }   
  else 
    {
      MPID_HANDLE_MUTEX_LOCK( shandle );
      shandle->is_cancelled = 0;
      MPID_HANDLE_MUTEX_UNLOCK( shandle );
      DEBUG_PRINT_MSG("Unable to cancel request");
    }
  MPID_HANDLE_CANCEL_READY( shandle );
  MPID_DECREMENT_CANCEL_ACK();  
} 
/*---------------------------------------------------------------------------------*/
/* This routine will block while a process is still expecting an ack from
   a cancel request.  Called by MPID_CH_End */
void MPID_FinishCancelPackets( void )
{  
  DEBUG_PRINT_MSG("Entering MPID_FinishCancelPackets");
  LOCK_CANCEL_ACK(); 
  while (expect_cancel_ack > 0) 
    {
      WAIT_CANCEL_ACK();
    } 
  UNLOCK_CANCEL_ACK();
  DEBUG_PRINT_MSG("Leaving MPID_FinishCancelPackets");
  
} 
/*---------------------------------------------------------------------------------*/










