/************************************************************/
#include "hot_stuff.h"
/************************************************************/
/*                 Prototype definitions                    */
/************************************************************/
int MPID_CH_Mad_Rndv_internal_send( void *buf,
				    int len,
				    int src_lrank,
				    int tag,
				    int context_id,
				    int dest,
				    MPID_Msgrep_t  msgrep,
				    MPIR_SHANDLE *shandle,
				    p_mad_channel_t channel );
/*-----------------------------------------------------------*/
int MPID_CH_Mad_Rndv_send( void *buf,
			   int len,
			   int src_lrank,
			   int tag,
			   int context_id,
			   int dest,
                           MPID_Msgrep_t msgrep,
			   MPIR_SHANDLE *shandle,
			   p_mad_channel_t channel );
/*-----------------------------------------------------------*/
int MPID_CH_Mad_Rndv_isend( void *buf,
			    int len,
			    int src_lrank,
			    int tag,
			    int context_id,
			    int dest, 
			    MPID_Msgrep_t msgrep, 
			    MPIR_SHANDLE *shandle, 
			    p_mad_channel_t channel );
/*------------------------------------------------------------*/
int MPID_CH_Mad_Rndv_save( MPIR_RHANDLE *rhandle,
			   int from,
			   void *in_pkt,
			   p_mad_channel_t channel );
/*------------------------------------------------------------*/
int MPID_CH_Mad_Rndv_ok_to_send( MPID_Aint send_id,
				 MPID_Aint recv_id,
				 int from,
				 p_mad_channel_t channel );
/*------------------------------------------------------------*/
int MPID_CH_Mad_Rndv_ack( void *in_pkt,
			  int from_grank,
			  p_mad_channel_t channel );
/*-------------------------------------------------------------*/
void MPID_CH_Mad_Rndv_tisend( thread_args *args );
/*-------------------------------------------------------------*/
int MPID_CH_Mad_Rndv_irecv( MPIR_RHANDLE *rhandle,
			    int from,
			    void *in_pkt,
			    p_mad_channel_t channel );
/*-------------------------------------------------------------*/
int MPID_CH_Mad_Rndv_unxrecv_start( MPIR_RHANDLE *rhandle, 
				    void *in_runex );
/*-------------------------------------------------------------*/
int MPID_CH_Mad_Rndv_unxrecv_end( MPIR_RHANDLE *rhandle );
/*-------------------------------------------------------------*/
/*int MPID_CH_Mad_Rndv_unxrecv_test_end( MPIR_RHANDLE *rhandle ); */

/*=============================================================*/
#ifndef MPID_HAS_HETERO
/* internal functions */
void MPID_CH_Mad_Typesend( p_mad_connection_t connection, 
			   void *buf, 
			   int count, 
			   struct MPIR_DATATYPE *dtype );
/*--------------------------------------------------------------*/
void MPID_CH_Mad_Typereceive( p_mad_connection_t connection, 
			      void *buf, 
			      int count, 
			      struct MPIR_DATATYPE *dtype );
/*---------------------------------------------------------------*/
/* public function */
int MPID_CH_Mad_Rndv_sendtype( void *buf,
			       int len, 
			       struct MPIR_DATATYPE *dtype, 
			       int src_lrank,
			       int tag,
			       int context_id,
			       int dest,
			       MPID_Msgrep_t msgrep,
			       p_mad_channel_t channel ); 
#endif /* HETERO */ 
/*================================================================*/
/*================================================================*/
#ifdef MPID_RNDV_SELF
int MPID_CH_Mad_Rndv_save_self( MPIR_RHANDLE *rhandle,
			        int from,
			        void *in_pkt );
/*----------------------------------------------------------------*/
int MPID_CH_Mad_Rndv_unxrecv_start_self( MPIR_RHANDLE *rhandle,
					 void *in_runex );
#endif /* MPID_RNDV_SELF */
/*================================================================*/
void MPID_CH_Mad_Rndv_delete( MPID_Protocol *p );
/*----------------------------------------------------------------*/

/******************************************************************/
/*                       End of Prototypes                        */
/******************************************************************/



/******************************************************************/
/*             Definitions of the actual functions                */
/**************************************************************** */
int MPID_CH_Mad_Rndv_internal_send(void *buf,
				   int len,
				   int src_lrank,
				   int tag,
				   int context_id,
				   int dest,
				   MPID_Msgrep_t  msgrep,
				   MPIR_SHANDLE *shandle,
				   p_mad_channel_t channel )

{
   p_mad_connection_t       connection;
   CH_MAD_HEADER_t          header __NET_BUFFER__;
   /* marcel_mutex_t           *mutex;*/
#ifdef GLOBAL_DEBUG
   marcel_t                 tid = marcel_self();
   
   marcel_fprintf(stderr," ----> Entering Rndv internal send ...\n");
   header.node_id                           = MPID_MyWorldRank; 
   header.tid                               = tid->number; 
#endif
   
   header.type                              = MAD_REQUEST_PKT;
   header.opt_field.request_pkt.mode        = MPID_PKT_REQUEST_SEND;
   header.opt_field.request_pkt.context_id  = context_id;
   header.opt_field.request_pkt.lrank	    = src_lrank;
   header.opt_field.request_pkt.to	    = dest; // Global rank
   header.opt_field.request_pkt.src         = MPID_MyWorldRank; 
   header.opt_field.request_pkt.tag	    = tag;
   

   
   MPID_HANDLE_MUTEX_LOCK( shandle );
   if (shandle->datatype)
     {  
	header.opt_field.request_pkt.len    = len*shandle->datatype->extent;
     }
   else
     {
	header.opt_field.request_pkt.len    = len;
     }   
   MPID_DO_HETERO(header.opt_field.request_pkt.msgrep = (int)msgrep);  
   MPID_AINT_SET(header.opt_field.request_pkt.send_id,shandle);
   
   /* Store info in the request for completing the message */
   MPID_HANDLE_COMPLETE_RESET( shandle );
   shandle->start	    = buf;
   shandle->bytes_as_contig = len; 
   /* Set the test/wait functions */
   shandle->wait	    = MPID_WaitForCompleteSend;
   shandle->test            = NULL;   /* default will work */
   shandle->finish          = NULL;
   shandle->channel         = channel;
   MPID_HANDLE_MUTEX_UNLOCK( shandle );
   
   /* Hetero stuff */
   MPID_PKT_PACK( &(header.opt_field.request_pkt), 
		  sizeof(header.opt_field.request_pkt),dest);
   MPID_PKT_LEN_SET(&(header.opt_field.request_pkt),
		    sizeof(header.opt_field.request_pkt));  
   MPID_INCREMENT_PENDING();
    
#ifdef GLOBAL_DEBUG
   marcel_printf("---> Rndv internal send : sending REQUEST PKT of size : \n 
		   (header %i) (header type %i) (packet %i) \n",sizeof(CH_MAD_HEADER_t)
		 ,header.type 
		 ,sizeof(MPID_PKT_REQUEST_SEND_T));
#endif 
   
   /*  mutex = Mutex_From_Channel( channel , MPID_MAD_Channel_Table );
    LOCK_CHANNEL( mutex ); */

   // Don't forget to reconvert the global rank to local one, for the channel.
   connection = mad_begin_packing( channel ,
				   MPID_CH_Mad_Global_to_Local( channel, dest ));
   mad_pack( connection,
	     &header,
	     SIZE_OF_MADHEADER,
	     mad_send_CHEAPER, 
	     mad_receive_EXPRESS); 
   
   mad_end_packing(connection);
   
   /*UNLOCK_CHANNEL( mutex ); */
#ifdef GLOBAL_DEBUG
   marcel_printf("<--- Rndv internal Send : done.\n");
#endif
   return MPI_SUCCESS;
}
/*--------------------------------------------------------------*/
int MPID_CH_Mad_Rndv_send(
			  void *buf,
			  int len,
			  int src_lrank,
			  int tag,
			  int context_id,
			  int dest,
                          MPID_Msgrep_t msgrep,
			  MPIR_SHANDLE *shandle,
			  p_mad_channel_t channel)
{
  return MPID_CH_Mad_Rndv_sendtype(buf, 
				   len, 
				   NULL, 
				   src_lrank, 
				   tag,
				   context_id, 
				   dest, 
				   msgrep,
				   channel );
}
/*--------------------------------------------------------------*/
int MPID_CH_Mad_Rndv_sendtype(
			      void *buf,
			      int count, 
			      struct MPIR_DATATYPE *type, 
			      int src_lrank,
			      int tag,
			      int context_id,
			      int dest,
			      MPID_Msgrep_t msgrep,
			      p_mad_channel_t channel )
{
  MPIR_SHANDLE *shandle;
  int err = MPI_SUCCESS;
  
#ifdef GLOBAL_DEBUG
  marcel_printf("===> Entering Send type ...\n");
#endif
  
  MPID_SendAlloc( shandle );
  MPID_Request_init( shandle, MPIR_SEND );
  
  MPID_HANDLE_MUTEX_LOCK( shandle );
  DEBUG_INIT_STRUCT( shandle, sizeof(MPIR_SHANDLE));
  MPIR_SET_COOKIE( shandle, MPIR_REQUEST_COOKIE);
  shandle->datatype = type;
  MPID_HANDLE_MUTEX_UNLOCK(  shandle );
  if (type)
    {
      MPIR_REF_INCR(type);
    }
  
  err = MPID_CH_Mad_Rndv_internal_send( buf,
				        count, 
				        src_lrank,
				        tag,
				        context_id,
				        dest,
				        msgrep,
				        shandle,
				        channel);
    
  (*(shandle->wait))( shandle ); 
  MPID_SendFree( shandle );
#ifdef GLOBAL_DEBUG
  marcel_printf("<===== Exiting Send Type\n");
#endif
  return err;
}
/*-------------------------------------------------------------*/
void  MPID_CH_Mad_Rndv_tisend(thread_args *args)
{
#ifdef GLOBAL_DEBUG
   marcel_t tid = marcel_self();
   marcel_printf("Temporary Tisend : %i\n", tid->number);
#endif /* GLOBAL_DEBUG */
   
   MPID_CH_Mad_Rndv_internal_send( args->buf, 
				  args->len, 
				  args->src_lrank, 
				  args->tag,
				  args->context_id, 
				  args->dest, 
				  args->msgrep,
				  args->shandle,
				  args->channel );
  
  FREE(args);
}
/*--------------------------------------------------------------*/
int MPID_CH_Mad_Rndv_isend( void *buf,
			    int len,
			    int src_lrank,
			    int tag,
			    int context_id,
			    int dest, 
			    MPID_Msgrep_t msgrep, 
			    MPIR_SHANDLE *shandle,
			    p_mad_channel_t channel )
{
  thread_args   *args;
  marcel_attr_t attributes; 
  int           mpi_errno = MPI_SUCCESS; 
     
#ifdef GLOBAL_DEBUG
  marcel_printf("----> Entering Rdndv Isend \n");
#endif
  MPID_HANDLE_MUTEX_LOCK( shandle );
  shandle->partner  = dest;
  MPID_HANDLE_MUTEX_UNLOCK( shandle ); 
  
  args              = NEW( thread_args );
  args->buf         = buf; 
  args->len         = len;
  args->src_lrank   = src_lrank;
  args->tag         = tag;
  args->context_id  = context_id;
  args->dest        = dest;
  args->msgrep      = msgrep;
  args->shandle     = shandle; 
  args->channel     = channel;

  marcel_attr_init(&attributes);	
  marcel_attr_setdetachstate(&attributes, TRUE); 
  marcel_create(NULL, &attributes,
		(marcel_func_t)MPID_CH_Mad_Rndv_tisend, args);
  
#ifdef GLOBAL_DEBUG
  marcel_printf("<---- Exiting Rdndv Isend\n");
#endif
  
  return mpi_errno; 
}
/*----------------------------------------------------------------*/
/*
 * This is the routine called when a packet of type 
 * MPID_PKT_REQUEST_SEND is seen and the receive has been posted.
 */
int MPID_CH_Mad_Rndv_irecv( MPIR_RHANDLE *rhandle,
			    int from,
			    void *in_pkt,
			    p_mad_channel_t channel )
{
  MPID_PKT_REQUEST_SEND_T *pkt     = (MPID_PKT_REQUEST_SEND_T *)in_pkt;
  MPID_Aint               send_id, recv_id;
  int                     err      = MPI_SUCCESS;
  int                     msglen;
#ifdef DEBUG_POLLING
  marcel_t                tid = marcel_self(); 
#endif   

#if defined(MPID_RNDV_SELF)
  MPIR_SHANDLE            *shandle = NULL;
#endif
  
#ifdef DEBUG_POLLING
  marcel_printf("Entering Rndv Irecv\n");   
#endif
  /* A request packet is a little larger than the basic packet size and 
     may need to be unpacked (in the heterogeneous case) */
  MPID_PKT_UNPACK( (MPID_PKT_HEAD_T *)in_pkt + 1, 
		   sizeof(MPID_PKT_REQUEST_SEND_T) - sizeof(MPID_PKT_HEAD_T),
		   from );
  msglen = pkt->len;

  /* Check for truncation */
  MPID_CHK_MSGLEN( rhandle, msglen, err);
  
  /* Note that if we truncate, We really must receive the message in two 
     parts; the part that we can store, and the part that we discard.
     This case is not yet handled. */

#ifdef DEBUG_POLLING
  marcel_printf("Locking mutex@(%p) ... ", &(rhandle->recv_handle.mutex));
  MPID_HANDLE_MUTEX_LOCK( rhandle );
  marcel_printf("Done.\n");
#else /* DEBUG POLLING */
  MPID_HANDLE_MUTEX_LOCK( rhandle );
#endif /* DEBUG POLLING */
  
  rhandle->s.count	  = msglen;
  rhandle->s.MPI_TAG	  = pkt->tag;   
  rhandle->s.MPI_SOURCE   = pkt->lrank;
  rhandle->s.MPI_ERROR    = err;
  
#ifdef DEBUG_POLLING
  marcel_printf("Signaling status ready ... ");
  MPID_HANDLE_STATUS_READY( rhandle ); /* syncro point */ 
  marcel_printf("Done.\n");
  marcel_printf("UNLocking mutex@(%p) ... ", &(rhandle->recv_handle.mutex));
  MPID_HANDLE_MUTEX_UNLOCK( rhandle );
  marcel_printf("Done.\n");
#else /* DEBUG POLLING */
  MPID_HANDLE_STATUS_READY( rhandle ); /* syncro point */ 
  MPID_HANDLE_MUTEX_UNLOCK( rhandle );  
#endif /* DEBUG POLLING */
  
#if defined(MPID_RNDV_SELF)
  if (from == MPID_MyWorldRank) 
    {
      /* first, we must find who we are ... */
      MPID_AINT_GET(shandle,pkt->send_id);
      MPID_HANDLE_MUTEX_LOCK( shandle );
#ifdef MPIR_HAS_COOKIES
      if (shandle->cookie != MPIR_REQUEST_COOKIE) 
	{
	  marcel_fprintf( stderr, "shandle is %lx\n", (long)shandle );
	  marcel_fprintf( stderr, "shandle cookie is %lx\n", shandle->cookie );
	  MPID_Print_shandle( stderr, shandle );
	  MPID_Abort( (struct MPIR_COMMUNICATOR *)0, 1, "MPI internal", 
		      "Bad address in Rendezvous send (irecv-self)" );
	}
#endif /* MPIR_HAS_COOKIES */
      /* Copy directly from the shandle */
      MEMCPY( rhandle->buf, shandle->start, shandle->bytes_as_contig );
      
      if(shandle->finish)
	{
	  (*(shandle->finish))( shandle );
	}
      MPID_HANDLE_MUTEX_UNLOCK( shandle );  
      MPID_HANDLE_COMPLETE_READY( shandle ); /* syncro point */     

      MPID_DECREMENT_PENDING();
      
      /* Update all of the rhandle information */
      rhandle->wait	 = NULL;
      rhandle->test	 = NULL;
      rhandle->push	 = NULL;
      if(rhandle->finish)
	{
	  (*(rhandle->finish))( rhandle );
	}
      MPID_HANDLE_MUTEX_UNLOCK( rhandle );  
      MPID_HANDLE_COMPLETE_READY( rhandle ); /* syncro point */          
      MPID_BREAK_LOOP();
      return err;
    }
#endif /* MPID_RNDV_SELF */
  
#ifdef DEBUG_POLLING
  marcel_printf("Locking mutex@(%p) ... ", &(rhandle->recv_handle.mutex));
  MPID_HANDLE_MUTEX_LOCK( rhandle );
  marcel_printf("Done.\n");
#else /* DEBUG POLLING */
  MPID_HANDLE_MUTEX_LOCK( rhandle );
#endif /* DEBUG POLLING */
  
  MPID_AINT_SET( rhandle->send_id , pkt->send_id ); /* very important  */
  MPID_AINT_SET( recv_id , rhandle );               /* local copy for the call */
  MPID_AINT_SET( send_id , rhandle->send_id );      /* local copy for the call */

#ifdef GLOBAL_DEBUG
  if(rhandle->datatype) 
    {
      marcel_printf("preparing to receive %p type\n", rhandle->datatype);
    }
#endif
  
  rhandle->wait	       = MPID_CH_Mad_Rndv_unxrecv_end;
  rhandle->test	       = NULL;
  rhandle->push	       = NULL;

#ifdef DEBUG_POLLING
  marcel_printf("Reseting rhandle is_complete ... ");
  MPID_HANDLE_COMPLETE_RESET( rhandle );
  marcel_printf("Done.\n");
  marcel_printf("UNLocking mutex@(%p) ... ", &(rhandle->recv_handle.mutex));
  MPID_HANDLE_MUTEX_UNLOCK( rhandle );
  marcel_printf("Done.\n");
  marcel_printf("[NODE #%i][THREAD #%i]\t ======> SENDING : LOCAL rhandle @ %p :\n",
		MPID_MyWorldRank,tid->number, recv_id);
  marcel_printf("Exiting Rndv Irecv.\n");
#else /* DEBUG POLLING */
  MPID_HANDLE_COMPLETE_RESET( rhandle );
  MPID_HANDLE_MUTEX_UNLOCK( rhandle );
#endif /* DEBUG POLLING */
  
#ifdef GLOBAL_DEBUG
  MPID_HANDLE_MUTEX_LOCK( rhandle );
  marcel_printf("---------- Rndv Irecv -----------\n");
  PRINT_RHANDLE( rhandle );
  marcel_printf("---------- Rndv Irecv -----------\n");
  MPID_HANDLE_MUTEX_UNLOCK( rhandle );
#endif
  
  MPID_CH_Mad_Rndv_ok_to_send( send_id, 
			       recv_id,
			       from,
			       channel );
  
  return err;
}
/*----------------------------------------------------------------------------*/
/* Save an unexpected message in rhandle */
int MPID_CH_Mad_Rndv_save( MPIR_RHANDLE *rhandle,
			   int from,
			   void *in_pkt,
			   p_mad_channel_t channel )
{
  MPID_PKT_REQUEST_SEND_T   *pkt = (MPID_PKT_REQUEST_SEND_T *)in_pkt;

#ifdef DEBUG_POLLING
  marcel_printf("Entering Rndv Save ...\n");
#endif
  /* A request packet is a little larger than the basic packet size and 
     may need to be unpacked (in the heterogeneous case) */
  MPID_PKT_UNPACK( (MPID_PKT_HEAD_T *)in_pkt + 1, 
		   sizeof(MPID_PKT_REQUEST_SEND_T) - sizeof(MPID_PKT_HEAD_T),
		   from );
  
#if defined(MPID_RNDV_SELF)
  if (from == MPID_MyWorldRank) 
    {
      return MPID_CH_Mad_Rndv_save_self( rhandle, from, in_pkt );
    }
#endif
  
#ifdef DEBUG_POLLING
  marcel_printf("Locking mutex@(%p) ... ", &(rhandle->recv_handle.mutex));
  MPID_HANDLE_MUTEX_LOCK( rhandle );
  marcel_printf("Done.\n");
#else /* DEBUG POLLING */
  MPID_HANDLE_MUTEX_LOCK( rhandle );
#endif /* DEBUG POLLING */

  rhandle->s.MPI_TAG	     = pkt->tag;
  rhandle->s.MPI_SOURCE      = pkt->lrank;
  rhandle->s.MPI_ERROR       = 0; /* MPI_SUCCESS */
  rhandle->s.count           = pkt->len;
   
#ifdef DEBUG_POLLING
  marcel_printf("Signaling Status ready ... ");
  MPID_HANDLE_STATUS_READY( rhandle ); /* syncro point */
  marcel_printf("Done.\n");
#else /* DEBUG POLLING */
  MPID_HANDLE_STATUS_READY( rhandle ); /* syncro point */
#endif /* DEBUG POLLING */

  rhandle->from              = from;
  rhandle->send_id           = pkt->send_id;

#ifdef DEBUG_POLLING
  marcel_printf("Reseting rhandle is_complete ... ");
  MPID_HANDLE_COMPLETE_RESET( rhandle );
  marcel_printf("Done.\n");
#else /* DEBUG POLLING */
  MPID_HANDLE_COMPLETE_RESET( rhandle );
#endif /* DEBUG POLLING */

  rhandle->channel           = channel;

  MPID_DO_HETERO(rhandle->msgrep = (MPID_Msgrep_t)pkt->msgrep );

  /* Need to set the push etc routine to complete this transfer */  
  rhandle->push              = MPID_CH_Mad_Rndv_unxrecv_start;

#ifdef DEBUG_POLLING
  marcel_printf("Signaling PUSH  ready ... ");
  MPID_HANDLE_PUSH_READY( rhandle ); /* syncro point */ 
  marcel_printf("Done.\n");
  marcel_printf("UNLocking mutex@(%p) ... ", &(rhandle->recv_handle.mutex));
  MPID_HANDLE_MUTEX_UNLOCK( rhandle );
  marcel_printf("Done.\n");
  marcel_printf("... Exiting Rndv_save\n");
#else /* DEBUG POLLING */
  MPID_HANDLE_PUSH_READY( rhandle ); /* syncro point */ 
  MPID_HANDLE_MUTEX_UNLOCK( rhandle );
#endif /* DEBUG POLLING */

#ifdef GLOBAL_DEBUG
  MPID_HANDLE_MUTEX_LOCK( rhandle );
  marcel_printf("---------- Rndv Save -----------\n");
  PRINT_RHANDLE( rhandle );
  marcel_printf("---------- Rndv Save -----------\n");
  MPID_HANDLE_MUTEX_UNLOCK( rhandle );
#endif
  
  return 0;
}
/*------------------------------------------------------------------------*/
int MPID_CH_Mad_Rndv_internal_ok_to_send( MPID_Aint send_id,
					  MPID_Aint recv_id,
					  int from,
					  p_mad_channel_t channel )
{
   p_mad_connection_t    connection;
   CH_MAD_HEADER_t       header __NET_BUFFER__;
   /*  marcel_mutex_t        *mutex;*/
#ifdef GLOBAL_DEBUG
   marcel_t              tid = marcel_self();
   header.node_id                   = MPID_MyWorldRank; 
   header.tid                       = tid->number; 
#endif /* GLOBAL_DEBUG */   
   
   header.type                      = MAD_SENDOK_PKT;
   header.opt_field.sendok_pkt.mode = MPID_PKT_OK_TO_SEND;
   header.opt_field.sendok_pkt.src  = MPID_MyWorldRank;
   MPID_AINT_SET( header.opt_field.sendok_pkt.send_id , send_id ); 
   MPID_AINT_SET( header.opt_field.sendok_pkt.recv_id , recv_id ); /* very important ... */
   MPID_PKT_PACK( &(header.opt_field.sendok_pkt),
		  sizeof(header.opt_field.sendok_pkt),
		  from );
   MPID_PKT_LEN_SET( &(header.opt_field.sendok_pkt) , 
		     sizeof(header.opt_field.sendok_pkt) );
   
#ifdef GLOBAL_DEBUG

   marcel_printf("[NODE #%i][THREAD #%i]\t sending OK TO SEND PKT : LOCAL RHANDLE is @ %p\n",
		 MPID_MyWorldRank, tid->number, header.opt_field.sendok_pkt.recv_id);
#endif
   
  /*  mutex = Mutex_From_Channel( channel , MPID_MAD_Channel_Table );
   LOCK_CHANNEL( mutex ); */
  
   connection = mad_begin_packing( channel , 
				   MPID_CH_Mad_Global_to_Local( channel, from ));
   mad_pack( connection,
	     &header,
	     SIZE_OF_MADHEADER,
	     mad_send_CHEAPER,
	     mad_receive_EXPRESS); 
   mad_end_packing(connection);
   /*  UNLOCK_CHANNEL( mutex ); */
   
#ifdef GLOBAL_DEBUG
   marcel_printf("<--- Internal OK to Send : Done.\n");
#endif
   return MPI_SUCCESS;
}
/*------------------------------------------------------------------------*/
void  MPID_CH_Mad_Rndv_tok_to_send( thread_args2 *args )
{
#ifdef GLOBAL_DEBUG
   marcel_t tid = marcel_self();
   marcel_printf("Temporary TOK_TO_SEND : %i\n", tid->number);
#endif
   
   MPID_CH_Mad_Rndv_internal_ok_to_send( args->send_id, 
					 args->recv_id, 
					 args->from,
					 args->channel );  
   FREE(args);
}
/*-------------------------------------------------------------------------*/
/*
 * This is an internal routine to return an OK TO SEND packet
 */
int MPID_CH_Mad_Rndv_ok_to_send( MPID_Aint send_id,
				 MPID_Aint recv_id,
				 int from,
				 p_mad_channel_t channel )
{
   thread_args2 *args;
   marcel_attr_t attributes;
   int           mpi_errno = MPI_SUCCESS; 
   
   args           = NEW( thread_args2 );
   args->send_id  = send_id; 
   args->recv_id  = recv_id;
   args->from     = from;
   args->channel  = channel;

   marcel_attr_init(&attributes);	
   marcel_attr_setdetachstate(&attributes, TRUE); 
   marcel_create( NULL, &attributes,
		  (marcel_func_t) MPID_CH_Mad_Rndv_tok_to_send, args);
   return mpi_errno; 
}
/*-------------------------------------------------------------------*/
/* 
 * This routine is called when it is time to receive an unexpected
 * message
 */
int MPID_CH_Mad_Rndv_unxrecv_start( MPIR_RHANDLE *rhandle,
				    void *in_runex )
{
  MPIR_RHANDLE    *runex = (MPIR_RHANDLE *)in_runex;
  MPID_Aint       send_id, recv_id;
  p_mad_channel_t channel;
  int             from;
  
  MPID_HANDLE_MUTEX_LOCK( runex );
  MPID_AINT_SET( send_id , runex->send_id );
  MPID_AINT_SET( recv_id , rhandle );  
                /* and not runex->recv_handle.recv_id ... */
  from                 = runex->from;
  channel              = runex->channel; 
  MPID_HANDLE_MUTEX_LOCK( rhandle );  
  rhandle->s	       = runex->s;
  rhandle->wait	       = MPID_CH_Mad_Rndv_unxrecv_end;
  rhandle->test	       = NULL; 
  rhandle->push	       = NULL;
  rhandle->from        = runex->from;
  rhandle->channel     = runex->channel; 
  MPID_HANDLE_MUTEX_UNLOCK( rhandle );
  MPID_HANDLE_MUTEX_UNLOCK( runex );  

#ifdef GLOBAL_DEBUG
  marcel_printf("--------> SWITCH done for %p and %p <------------\n", 
		runex, rhandle);
  MPID_HANDLE_MUTEX_LOCK( rhandle );
  marcel_printf("---------- Rndv Unex Start -----------\n");
  PRINT_RHANDLE( rhandle );
  marcel_printf("---------- Rndv Unex Start -----------\n");
  MPID_HANDLE_MUTEX_UNLOCK( rhandle );
#endif

  MPID_RecvFree( runex );

  /* this should be the last thing to do */
  MPID_CH_Mad_Rndv_ok_to_send( send_id,
			       recv_id,
			       from,
			       channel );
  return 0;
}
/*---------------------------------------------------------------------*/
/* 
   This is the wait routine for a rendezvous message that was unexpected.
   A request for the message has already been sent.
   */
int MPID_CH_Mad_Rndv_unxrecv_end( MPIR_RHANDLE *rhandle )
{
  return MPID_WaitForCompleteRecv( rhandle ); 
}
/*---------------------------------------------------------------------*/
/* 
   This is the test routine for a rendezvous message that was unexpected.
   A request for the message has already been sent.
   */
/* 
   int MPID_CH_Mad_Rndv_unxrecv_test_end(MPIR_RHANDLE *rhandle) 
   {
   return 0; 
   }
   */

/*----------------------------------------------------------------------*/
/* This is the routine that is called when an "ok to send" packet is
   received. */
int MPID_CH_Mad_Rndv_internal_ack( void *in_pkt,
				   int from_grank,
				   p_mad_channel_t  channel )
{
  p_mad_connection_t    connection;
  CH_MAD_HEADER_t       header    __NET_BUFFER__;
  MPID_PKT_OK_TO_SEND_T *pkt      = (MPID_PKT_OK_TO_SEND_T *)in_pkt; 
  MPIR_SHANDLE          *shandle  = NULL;
  /*marcel_mutex_t        *mutex; */
#ifdef GLOBAL_DEBUG
   marcel_t              tid = marcel_self(); 
   
   marcel_printf("[NODE #%i][THREAD #%i]\t <============== ACKNOWLEDGING : REMOTE RHANDLE @ : %p\n",
		 MPID_MyWorldRank,tid->number,pkt->recv_id);    
#endif   
  /* A request packet is a little larger than the basic packet size and 
     may need to be unpacked (in the heterogeneous case) */
  MPID_PKT_UNPACK( (MPID_PKT_HEAD_T *)in_pkt + 1, 
		   sizeof(MPID_PKT_OK_TO_SEND_T) - sizeof(MPID_PKT_HEAD_T),
		   from_grank );
  
  /* we must find who handles the communication ... */
  MPID_AINT_GET( shandle , pkt->send_id );

   
#ifdef MPIR_HAS_COOKIES
  MPID_HANDLE_MUTEX_LOCK( shandle );
  if (shandle->cookie != MPIR_REQUEST_COOKIE) {
    marcel_fprintf( stderr, "shandle is %lx\n", (long)shandle );
    marcel_fprintf( stderr, "shandle cookie is %lx\n", shandle->cookie );
    MPID_Print_shandle( stderr, shandle );
    MPID_HANDLE_MUTEX_UNLOCK( shandle );
    MPID_Abort( (struct MPIR_COMMUNICATOR *)0, 1, "MPI internal", 
		"Bad address in Rendezvous send (ack)" );
  }
  else
    {
      MPID_HANDLE_MUTEX_UNLOCK( shandle );
    }
#endif /* MPIR_HAS_COOKIES */
  
#ifdef GLOBAL_DEBUG
  MPID_HANDLE_MUTEX_LOCK( shandle );
  if(shandle->channel != channel)
    {
      MPID_HANDLE_MUTEX_UNLOCK( shandle );
      MPID_Abort( (struct MPIR_COMMUNICATOR *)0, 1, "MPI internal",
		  "Channel mismatch in rndv send (ack)\n");
    }
  else
    {
      MPID_HANDLE_MUTEX_UNLOCK( shandle );
    }
#endif /* GLOBAL_DEBUG */
  
#ifdef MPID_DEBUG_ALL
  if (MPID_DebugFlag) 
    {
      marcel_fprintf( MPID_DEBUG_FILE, "[%d]S for ", MPID_MyWorldRank );
      MPID_Print_shandle( MPID_DEBUG_FILE, shandle );
    }
#endif /* MPID_DEBUG_ALL */
  
  MPID_DECREMENT_PENDING();
  
  header.type                      = MAD_RNDV_PKT; 
  MPID_AINT_SET( header.opt_field.rhandle_address , pkt->recv_id );
   
#ifdef GLOBAL_DEBUG
  header.node_id                   = MPID_MyWorldRank; 
  header.tid                       = tid->number;  
   
  marcel_printf("[NODE #%i][THREAD #%i]\t <============== ACKNOWLEDGING (2): REMOTE RHANDLE @ : %p\n",
		 MPID_MyWorldRank,tid->number, header.opt_field.rhandle_address);    
  marcel_printf("[NODE #%i][THREAD #%i]\t <============== ACKNOWLEDGING (3): REMOTE RHANDLE @ : %p\n",
		 MPID_MyWorldRank,tid->number, pkt->recv_id);    
  marcel_printf("---> Rndv internal ack : sending RNDV PKT header\n
                (header type %i) (rhandle @ %p) \n"
		,header.type
		,header.opt_field.rhandle_address);
#endif /* GLOBAL_DEBUG */

  /*   mutex = Mutex_From_Channel( channel , MPID_MAD_Channel_Table );
       LOCK_CHANNEL( mutex ); */
  connection = mad_begin_packing( channel ,
				  MPID_CH_Mad_Global_to_Local( channel, 
							       from_grank ));
#ifdef DEBUG_POLLING
  marcel_printf("Internal Ack Sending Data on channel\n ");
#endif
  
  mad_pack( connection, 
	    &header, 
	    SIZE_OF_MADHEADER, 
	    mad_send_CHEAPER, 
	    mad_receive_EXPRESS ); 

#ifdef GLOBAL_DEBUG  
  marcel_printf("<--- Rndv PKT header sent.\n");
#endif /* GLOBAL_DEBUG */
  MPID_HANDLE_MUTEX_LOCK( shandle );
  
#ifndef MPID_HAS_HETERO
  if(!(shandle->datatype))
    {
      if (shandle->bytes_as_contig > 0)
	{
#ifdef ENFORCE_ALIGNMENT
	  int size = tbx_aligned( shandle->bytes_as_contig , MAD_LENGTH_ALIGNMENT ); 
#else /* ENFORCE_ALIGNMENT */
	  int size = shandle->bytes_as_contig;
#endif /* ENFORCE_ALIGNMENT */ 	  
	  
#ifdef GLOBAL_DEBUG
	  marcel_printf("---> Rndv internal ack : sending buffer (size %i)\n", shandle->bytes_as_contig);
#endif	  
	  mad_pack( connection,
		    shandle->start,
		    size,
		    mad_send_CHEAPER,
		    mad_receive_CHEAPER );
#ifdef GLOBAL_DEBUG
	  marcel_printf("<--- Rndv internal ack : Buffer sent.\n");
#endif
	}
    }
  else
    {
#ifndef ENFORCE_ALIGNMENT   
      if (shandle->datatype->extent >=  TYPE_SWITCH) 
	{   
	  
#ifdef GLOBAL_DEBUG 
	  marcel_printf("Rndv Ack : Sending typed rndv (%i itms)\n", shandle->bytes_as_contig); 
#endif	    
	  MPID_CH_Mad_Typesend( connection, 
			        shandle->start, 
			        shandle->bytes_as_contig, 
			        shandle->datatype );	
	} 
      else
#endif /* ENFORCE_ALIGNMENT */
	{ 
#ifdef ENFORCE_ALIGNMENT
	  int size = tbx_aligned( shandle->bytes_as_contig , MAD_LENGTH_ALIGNMENT ); 
#else /* ENFORCE_ALIGNMENT */
	  int size = shandle->bytes_as_contig;
#endif /* ENFORCE_ALIGNMENT */ 
	  mad_pack( connection,
		    shandle->start,
		    size,
		    mad_send_CHEAPER,
		    mad_receive_CHEAPER);
	  
	}
    }
#else /*  HETERO  */
  if (shandle->bytes_as_contig > 0)
    {
#ifdef ENFORCE_ALIGNMENT
      int size = tbx_aligned( shandle->bytes_as_contig , MAD_LENGTH_ALIGNMENT ); 
#else /* ENFORCE_ALIGNMENT */
      int size = shandle->bytes_as_contig;
#endif /* ENFORCE_ALIGNMENT */ 
      
      mad_pack( connection,
		shandle->start,
		size,
		mad_send_CHEAPER,
		mad_receive_CHEAPER);
    }
#endif  /*  HETERO */
  
  mad_end_packing(connection);
  /*  UNLOCK_CHANNEL( mutex );*/
  
#ifdef GLOBAL_DEBUG
   marcel_printf("[NODE #%i][THREAD #%i]\t <============== ACKNOWLEDGING (4): REMOTE RHANDLE @ : %p\n",
		 MPID_MyWorldRank,tid->number, header.opt_field.rhandle_address);    
   marcel_printf("[NODE #%i][THREAD #%i]\t Completed sending data on channel , RHANDLE WAS @ %p\n",
		 MPID_MyWorldRank,tid->number, pkt->recv_id);       
#endif  
   
  if(shandle->finish)
    (*(shandle->finish))( shandle );
  MPID_HANDLE_MUTEX_UNLOCK( shandle );
  
  MPID_HANDLE_COMPLETE_READY( shandle ); /* syncro point */

  MPID_BREAK_LOOP();

  OUTPUT("... Exiting Rndv internal ack");
  return MPI_SUCCESS;
}
/*------------------------------------------------------------------------*/
void  MPID_CH_Mad_Rndv_tack(thread_args3 *args)
{
/*  marcel_t tid = marcel_self();
  marcel_printf("Temporary RNDV ACK : %i\n", tid->number);
 */  
  MPID_CH_Mad_Rndv_internal_ack( &(args->in_pkt), 
				 args->from_grank,
				 args->channel ); 
  
  FREE(args);
}
/*-------------------------------------------------------------------------*/
int MPID_CH_Mad_Rndv_ack( void *in_pkt,
			  int from_grank,
			  p_mad_channel_t channel )
{
  thread_args3  *args;
  marcel_attr_t attributes;
  int           mpi_errno = MPI_SUCCESS; 
#ifdef GLOBAL_DEBUG  
  marcel_t      tid = marcel_self();  
  MPID_PKT_OK_TO_SEND_T *pkt =  (MPID_PKT_OK_TO_SEND_T *)in_pkt;
   
  marcel_printf("[NODE #%i][THREAD #%i]\t ACKNOWLEDGING an OK2SEND pkt ; REMOTE RHANDLE @ : %p\n",
		 MPID_MyWorldRank,tid->number,pkt->recv_id); 
#endif   
  args              = NEW( thread_args3 );
  MEMCPY( &(args->in_pkt.sendok_pkt) , (MPID_PKT_OK_TO_SEND_T *) in_pkt, sizeof(CH_MAD_PKT_t)); 
  args->from_grank  = from_grank;
  args->channel     = channel;
  
  marcel_attr_init(&attributes);	
  marcel_attr_setdetachstate(&attributes, TRUE);    
  marcel_create(NULL, &attributes, 
		(marcel_func_t) MPID_CH_Mad_Rndv_tack, args);

  return mpi_errno; 
}
/*-------------------------------------------------------------------------*/
void MPID_CH_Mad_Rndv_Handle( p_mad_connection_t connection, 
			      MPID_Aint rhandle_address )
{
  MPIR_RHANDLE *rhandle;

#ifdef DEBUG_POLLING
  marcel_t      tid = marcel_self();
   
  MPID_AINT_GET(rhandle, rhandle_address);
  marcel_printf("[NODE #%i][THREAD #%i]\t --------> Accessing RHANDLE @ : %p\n",
		 MPID_MyWorldRank,tid->number,rhandle); 
   if (!rhandle)
    {
      marcel_printf("NULL rhandle !!!! (%p) \n", (void *)rhandle_address );
    }
   
  marcel_printf(" Locking mutex@(%p) ... ", &(rhandle->recv_handle.mutex));   
  MPID_HANDLE_MUTEX_LOCK( rhandle );  
  marcel_printf("Done.\n");
#else /* DEBUG POLLING */
  MPID_AINT_GET(rhandle, rhandle_address);
  MPID_HANDLE_MUTEX_LOCK( rhandle );  
#endif /* DEBUG POLLING */

   
#ifndef MPID_HAS_HETERO  
  if(!rhandle->datatype)  
    { 
      if (rhandle->s.count > 0)
	{ 
#ifdef ENFORCE_ALIGNMENT
	  int size = tbx_aligned( rhandle->s.count, MAD_LENGTH_ALIGNMENT ); 
#else /* ENFORCE_ALIGNMENT */
	  int size = rhandle->s.count; 
#endif /* ENFORCE_ALIGNMENT */

#ifdef DEBUG_POLLING
	  marcel_printf("-------------> Rndv Handle : receiving msg size of %i\n", rhandle->s.count);   
#endif
	  mad_unpack ( connection, 
		       rhandle->buf, 
		       size,
		       mad_send_CHEAPER, 
		       mad_receive_CHEAPER);
#ifdef DEBUG_POLLING
	  marcel_printf("------------->... Rndv Handle : DONE\n"); 
#endif
	}
    } 
  else 
    { 
#ifndef ENFORCE_ALIGNMENT
      if (rhandle->datatype->extent >=  TYPE_SWITCH)  
	{ 
#ifdef DEBUG_POLLING 
	  marcel_printf("Rdv Handle : Receiving typed rndv (%i itms)\n", rhandle->count); 
#endif 
	  MPID_CH_Mad_Typereceive( connection, 
				   rhandle->start,
				   rhandle->count, 
				   rhandle->datatype ); 
	  
	}  
      else  
#endif /* ENFORCE_ALIGNMENT */
	{ 
#ifdef ENFORCE_ALIGNMENT
	  int size = tbx_aligned( rhandle->s.count, MAD_LENGTH_ALIGNMENT ); 
#else /* ENFORCE_ALIGNMENT */
	  int size = rhandle->s.count; 
#endif /* ENFORCE_ALIGNMENT */
	  
	  mad_unpack ( connection, 
		       rhandle->buf, 
		       size,
		       mad_send_CHEAPER, 
		       mad_receive_CHEAPER );
      	} 
    } 
#else /* HETERO */
  if (rhandle->s.count > 0)
    { 
#ifdef ENFORCE_ALIGNMENT
      int size = tbx_aligned( rhandle->s.count,MAD_LENGTH_ALIGNMENT ); 
#else /* ENFORCE_ALIGNMENT */
      int size = rhandle->s.count; 
#endif /* ENFORCE_ALIGNMENT */
      
      mad_unpack ( connection, 
		   rhandle->buf, 
		   size,
		   mad_send_CHEAPER, 
		   mad_receive_CHEAPER );
    }
#endif/* HETERO */
  
#ifdef DEBUG_POLLING
  marcel_printf("-----------------> End unpacking ...");
  mad_end_unpacking (connection);
  marcel_printf("Done.\n");
  
  if(rhandle->finish)
    {
      (*(rhandle->finish))( rhandle );
    }
  marcel_printf("----------------> UNLocking mutex@(%p) ... ", &(rhandle->recv_handle.mutex)); 
  MPID_HANDLE_MUTEX_UNLOCK( rhandle );  
  marcel_printf("Done.\n");
  
  marcel_printf("----------------> Signaling complete ready ...");
  MPID_HANDLE_COMPLETE_READY( rhandle ); /* syncro point */
  marcel_printf("Done.\n");
#else  /* DEBUG POLLING */
  mad_end_unpacking (connection);
  if(rhandle->finish)
    {  
      (*(rhandle->finish))( rhandle );
    }
  MPID_HANDLE_MUTEX_UNLOCK( rhandle );  
  MPID_HANDLE_COMPLETE_READY( rhandle ); /* syncro point */
#endif /* DEBUG POLLING */
  
  MPID_BREAK_LOOP();

#ifdef GLOBAL_DEBUG
  MPID_HANDLE_MUTEX_LOCK( rhandle );  
  marcel_printf("---------- Rndv Handle -----------\n");
  PRINT_RHANDLE( rhandle );
  marcel_printf("---------- Rndv Handle -----------\n");
  MPID_HANDLE_MUTEX_UNLOCK( rhandle );  
#endif /* GLOBAL_DEBUG */
}
/********************************************************************/
/*----------------------------------------------------------------*/
#if !defined( ENFORCE_ALIGNMENT) && !defined( MPID_HAS_HETERO )
void transmit_datatype( p_mad_connection_t connection, 
		        MPID_Aint buf, 
		        int count, 
		        struct MPIR_DATATYPE *dtype, 
		        int way )
{
  int i, j;
  MPID_Aint tmpbuf = NULL;
  
  if(dtype->is_contig)
    {
      if(way)
	{
#ifdef  GLOBAL_DEBUG
	  marcel_printf("Transmit datatype : packing  %i bytes \n", size);
#endif
	  mad_pack( connection, 
		    buf, 
		    count * dtype->extent,
		    mad_send_CHEAPER, 
		    mad_receive_CHEAPER );
	}
      else
	{
	  
#ifdef  GLOBAL_DEBUG
	  marcel_printf("Transmit datatype : unpacking  %i bytes \n", size);
#endif

	  mad_unpack( connection, 
		      buf, 
		      count * dtype->extent,
		      mad_send_CHEAPER, 
		      mad_receive_CHEAPER );
	}
      return;
    }
  
  switch (dtype->dte_type)
    {
      
    case MPIR_CONTIG:
      /* we could avoid this and call mad_pack directly */
      transmit_datatype(connection, buf, count * dtype->count, 
			dtype->old_type, way);
      break;
      
    case MPIR_HVECTOR: 
      for (i = 0; i < count; i++)
	{
	  tmpbuf = buf;
	  for (j = 0; j < dtype->count; j++)
	    {
	      transmit_datatype(connection, tmpbuf, 
				dtype->blocklen, dtype->old_type, way);
	      tmpbuf += (dtype->stride);
	    }
	  buf += dtype->extent;
	}
      break;
      
    case MPIR_VECTOR:
      for (i = 0; i < count; i++)
	{
	  tmpbuf = buf;
	  for (j = 0; j < dtype->count; j++)
	    {
	      transmit_datatype(connection, tmpbuf, 
				dtype->blocklen, dtype->old_type, way);
	      tmpbuf += ((dtype->stride)*(dtype->old_type->extent)); 
	    }   		  
	  buf += dtype->extent;
	}
      break;
      
    case MPIR_HINDEXED:  
      for (i = 0; i < count; i++)
	{
	  for (j = 0; j < dtype->count; j++)
	    {
	      tmpbuf  = buf + dtype->indices[j];
	      transmit_datatype(connection, tmpbuf, dtype->blocklens[j], 
				dtype->old_type, way);
	    }
	  buf += dtype->extent;
	}
      break;
      
    case MPIR_INDEXED: 
      for (i = 0; i < count; i++)
	{
	  for (j = 0; j < dtype->count; j++)
	    {
	      tmpbuf  = buf + ((dtype->indices[j])*(dtype->old_type->extent)); 
	      transmit_datatype(connection, tmpbuf, 
				dtype->blocklens[j], dtype->old_type, way);
	    }
	  buf += dtype->extent;
	}
      break;
      
    case MPIR_STRUCT:
      for (i = 0; i < count; i++)
	{
	  for (j = 0; j < dtype->count; j++)
	    {
	      tmpbuf  = buf + dtype->indices[j];
	      transmit_datatype(connection, tmpbuf, dtype->blocklens[j], 
				dtype->old_types[j], way);
	    }
	  buf  += dtype->extent;
	}
      break;
      
    default:
      marcel_printf("transmit: bad type!!\n");
      MPID_Abort( (struct MPIR_COMMUNICATOR *)0, 1, "MPI internal", 
		  "Transmit Datatype" );
    }
}
/*-------------------------------------------------------------------------*/
void MPID_CH_Mad_Typereceive( p_mad_connection_t connection, 
			      void *buf, 
			      int count, 
			      struct MPIR_DATATYPE *dtype )
{
  transmit_datatype(connection, buf, count, dtype, 0);
}
/*--------------------------------------------------------------------------*/
void MPID_CH_Mad_Typesend( p_mad_connection_t connection, 
			   void *buf, 
			   int count, 
			   struct MPIR_DATATYPE *dtype )
{
  transmit_datatype(connection, buf, count, dtype, 1);
}
#endif /* !HETERO && ! ALIGNMENT */
/***************************************************************************/
#if defined(MPID_RNDV_SELF)
/***************************************************************************/
/* This code handles rendezous messages to self                           
 */
/***************************************************************************/

/* Save an unexpected message in rhandle for sent to self */
int MPID_CH_Mad_Rndv_save_self(
			       MPIR_RHANDLE *rhandle,
			       int from,
			       void *in_pkt
			       )
{
  MPID_PKT_REQUEST_SEND_T   *pkt = (MPID_PKT_REQUEST_SEND_T *)in_pkt;
  
  /* A request packet is a little larger than the basic packet size and 
     may need to be unpacked (in the heterogeneous case) */
  MPID_PKT_UNPACK( (MPID_PKT_HEAD_T *)in_pkt + 1, 
		   sizeof(MPID_PKT_REQUEST_SEND_T) - sizeof(MPID_PKT_HEAD_T),
		   from );

  MPID_HANDLE_MUTEX_LOCK( rhandle );
  rhandle->s.MPI_TAG	= pkt->tag;
  rhandle->s.MPI_SOURCE = pkt->lrank;
  rhandle->s.MPI_ERROR  = 0;
  rhandle->s.count      = pkt->len;
  MPID_HANDLE_STATUS_READY( rhandle ); /* syncro point */
  MPID_HANDLE_COMPLETE_RESET( rhandle );
  rhandle->from         = from;
  rhandle->send_id      = pkt->send_id;
  
  /* Note that the send_id is just the address of the sending handle
     IN OUR ADDRESS SPACE */
  /* Note that format MIGHT be different from native (e.g., packed data) */
  MPID_DO_HETERO(rhandle->msgrep = (MPID_Msgrep_t)pkt->msgrep );
  
 /* Need to set the push etc routine to complete this transfer */
  rhandle->push         = MPID_CH_Mad_Rndv_unxrecv_start_self;
  MPID_HANDLE_PUSH_READY( rhandle ); /* syncro point */
  MPID_HANDLE_MUTEX_UNLOCK( rhandle );
  
  return MPI_SUCCESS ; 
}
/*----------------------------------------------------------------------------*/
/* 
 * This routine is called when it is time to receive an unexpected
 * message.  This is simple; we can just copy the data with memcpy.
 * Once the memcpy is done, we mark the SEND as completed.
 */
int MPID_CH_Mad_Rndv_unxrecv_start_self(
					MPIR_RHANDLE *rhandle,
					void *in_runex
					)
{
  MPIR_RHANDLE *runex = (MPIR_RHANDLE *)in_runex;
  MPIR_SHANDLE *shandle;
  
  /* Get the source handle */

  MPID_AINT_GET(shandle,runex->send_id);
  
#ifdef MPIR_HAS_COOKIES
  if (shandle->cookie != MPIR_REQUEST_COOKIE) 
    {
      marcel_fprintf( stderr, "shandle is %lx\n", (long)shandle );
      marcel_fprintf( stderr, "shandle cookie is %lx\n", shandle->cookie );
      MPID_Print_shandle( stderr, shandle );
      MPID_Abort( (struct MPIR_COMMUNICATOR *)0, 1, "MPI internal", 
		  "Bad address in Rendezvous send (unx_start_self)" );
    }
#endif /* MPIR_HAS_COOKIES */	
  
  MPID_HANDLE_MUTEX_LOCK( rhandle );
  MPID_HANDLE_MUTEX_LOCK( shandle );
  /* Copy directly from the shandle */
  MEMCPY( rhandle->buf, shandle->start, shandle->bytes_as_contig );
  if(shandle->finish)
    {  
      (*(shandle->finish))( shandle );
    }
  MPID_HANDLE_MUTEX_UNLOCK( shandle );
  MPID_HANDLE_COMPLETE_READY( shandle ); /* syncro point */

  MPID_DECREMENT_PENDING();
  
  /* Update all of the rhandle information */
  rhandle->wait	       = NULL;
  rhandle->test	       = NULL;
  rhandle->push	       = NULL;
  MPID_HANDLE_MUTEX_LOCK( runex );
  rhandle->s	       = runex->s;  
  MPID_HANDLE_STATUS_READY( rhandle ); /* syncro point */
  rhandle->from        = runex->from;
  MPID_HANDLE_MUTEX_UNLOCK( runex );
  if(rhandle->finish)
    {
      (*(rhandle->finish))( rhandle );
    }
  MPID_HANDLE_MUTEX_UNLOCK( rhandle );
  MPID_HANDLE_COMPLETE_READY( rhandle ); /* syncro point */
  
  BROADCAST_LOOP();
  
  MPID_RecvFree( runex );

  return MPI_SUCCESS;
}

/***************************************************************************/
#endif /* MPID_RNDV_SELF */

void MPID_CH_Mad_Rndv_delete(MPID_Protocol *p)
{
  FREE( p );
}
/*-------------------------------------------------------------------------*/
/*
 * The only routing really visable outside this file; it defines the
 * Blocking Rendezvous protocol.
 */
MPID_Protocol *MPID_CH_Mad_Rndv_setup(void)
{
    MPID_Protocol *p;

    p = NEW( MPID_Protocol );
    if (!p) 
      {
	return NULL;
      }
   
    p->send	   = MPID_CH_Mad_Rndv_send;
    p->recv	   = NULL;
    p->isend	   = MPID_CH_Mad_Rndv_isend;
    p->wait_send   = NULL;
    p->push_send   = NULL;
    p->cancel_send = NULL;
    p->irecv	   = MPID_CH_Mad_Rndv_irecv;
    p->wait_recv   = NULL;
    p->push_recv   = NULL;
    p->cancel_recv = NULL;
    p->do_ack      = MPID_CH_Mad_Rndv_ack;
    p->unex        = MPID_CH_Mad_Rndv_save;
    p->delete      = MPID_CH_Mad_Rndv_delete;

    return p;
}















