/*******************************************************************/
#include "hot_stuff.h"
/*******************************************************************/
/*                         PROTOTYPES                              */
/*******************************************************************/
int MPID_CH_Mad_eager_send( void * buf,
			    int len,
			    int src_lrank,
			    int tag,
                            int context_id,
			    int dest,
			    MPID_Msgrep_t msgrep,
			    MPIR_SHANDLE *shandle,
			    p_mad_channel_t  channel);
/*-----------------------------------------------------------------*/
int MPID_CH_Mad_eager_isend( void *buf,
			     int len,
			     int src_lrank,
			     int tag,
			     int context_id,
                             int dest, 
			     MPID_Msgrep_t msgrep, 
			     MPIR_SHANDLE *shandle, 
			     p_mad_channel_t channel);
/*-----------------------------------------------------------------*/
int MPID_CH_Mad_eager_save( MPIR_RHANDLE *rhandle,
			    int from, 
			    void *in_pkt, 
			    p_mad_channel_t channel);
/*------------------------------------------------------------------*/
void MPID_CH_Mad_eager_tisend( thread_args *args );
/*------------------------------------------------------------------*/
int MPID_CH_Mad_eager_recv( MPIR_RHANDLE *rhandle,
			    int from_grank,
			    void *in_pkt );
/*------------------------------------------------------------------*/
int MPID_CH_MAD_eager_unxrecv_start( MPIR_RHANDLE *rhandle,
				     void *in_runex );
/*------------------------------------------------------------------*/
void MPID_CH_Mad_delete( MPID_Protocol *p );
/*------------------------------------------------------------------*/
/********************************************************************/
/*                          DEFINITIONS                             */
/********************************************************************/
int MPID_CH_Mad_eager_send( void *buf,
			    int len,
			    int src_lrank,
			    int tag,
			    int context_id,
                            int dest,
			    MPID_Msgrep_t msgrep, 
			    MPIR_SHANDLE *shandle,
			    p_mad_channel_t channel)
{
   p_mad_connection_t connection;
   CH_MAD_HEADER_t    header __NET_BUFFER__;
   /*  marcel_mutex_t     *mutex; */
   
#ifdef GLOBAL_DEBUG 
   marcel_printf("--------> Entering short eager send for sending msg of length %i \n",len);
#endif 
   
   /* 
    These references are ordered to match the 
    order they appear in the structure 
    */
   header.type                          = MAD_SHORT_PKT;
   header.opt_field.head_pkt.mode       = MPID_PKT_SHORT; 
   header.opt_field.head_pkt.context_id = context_id;
   header.opt_field.head_pkt.lrank      = src_lrank;
   header.opt_field.head_pkt.to         = dest;
   header.opt_field.head_pkt.src        = MPID_MyWorldRank; 
   header.opt_field.head_pkt.tag        = tag;
   header.opt_field.head_pkt.len        = len; /* unaligned size of DATA */
   MPID_AINT_SET( header.opt_field.head_pkt.send_id , shandle );    
   
   /**** hetero stuff ****/
   MPID_DO_HETERO( header.opt_field.head_pkt.msgrep = (int)msgrep);
   MPID_PKT_PACK(&(header.opt_field.head_pkt), sizeof(MPID_PKT_HEAD_T), dest); 
   /*********************/  
   /*  mutex = Mutex_From_Channel( channel , MPID_MAD_Channel_Table );
    LOCK_CHANNEL( mutex ); */

   // Reverse the index before sending anything.
   connection = mad_begin_packing( channel , 
				    MPID_CH_Mad_Global_to_Local( channel, dest ));
#ifdef OPT
   if ( ( len > 0 ) && (len <= OPT_BUF_SIZE ))
     {	
	MEMCPY( header.padding, buf, len );
     }
#endif   
   mad_pack( connection, 
	     &header, 
	     SIZE_OF_MADHEADER,
	     mad_send_CHEAPER,
	     mad_receive_EXPRESS); 
   
   /*  marcel_printf("====> SEND: Header sent \n");*/
#ifdef OPT
   if (len > OPT_BUF_SIZE)
#else
   if ( len > 0)
#endif /* OPT */
       {
#ifdef ENFORCE_ALIGNMENT
	  int _len  = tbx_aligned( len , MAD_LENGTH_ALIGNMENT );
#else /* ENFORCE_ALIGNMENT */
	  int _len  = len;
#endif /* ENFORCE_ALIGNMENT */
	  
#ifdef GLOBAL_DEBUG
	  marcel_printf("=========> Eager Send : Sending %i \n", _len);
#endif 
	  mad_pack( connection, 
		    buf, 
		    _len, 
		    mad_send_CHEAPER, 
		    mad_receive_CHEAPER);  
       }
   mad_end_packing(connection);
   /* UNLOCK_CHANNEL( mutex );*/
   
#ifdef GLOBAL_DEBUG
   marcel_printf("<-------- Exiting short eager send .\n");
#endif   
   return MPI_SUCCESS;
}
/*----------------------------------------------------------------*/
void  MPID_CH_Mad_eager_tisend( thread_args *args )
{
  MPIR_SHANDLE *shandle = (MPIR_SHANDLE *)(args->shandle);

  //marcel_detach(marcel_self());
  
  MPID_CH_Mad_eager_send(args->buf,
			 args->len, 
			 args->src_lrank, 
			 args->tag,
			 args->context_id, 
			 args->dest, 
			 args->msgrep,
			 args->shandle,
			 args->channel);
  
  MPID_HANDLE_MUTEX_LOCK( shandle );
  if (shandle->finish)
    {
      (*(shandle->finish))(shandle);
    }
  MPID_HANDLE_MUTEX_UNLOCK( shandle );
  
  MPID_HANDLE_COMPLETE_READY( shandle ); /* syncro point */
#ifdef GLOBAL_DEBUG
  marcel_printf("Complete Ready for shandle @ %p\n", shandle );
#endif
  
  MPID_BREAK_LOOP(); 
  FREE( args ); 
}
/*----------------------------------------------------------------*/
int MPID_CH_Mad_eager_isend( void            *buf,
			     int             len,
			     int             src_lrank,
			     int             tag,
			     int             context_id,
                             int             dest, 
			     MPID_Msgrep_t   msgrep, 
			     MPIR_SHANDLE    *shandle,
			     p_mad_channel_t channel)
{
  thread_args *args;
  marcel_attr_t attributes; 
  int mpi_errno = MPI_SUCCESS; 
  
#ifdef GLOBAL_DEBUG
  marcel_printf(" ***** Using short isend for sending msg **** \n");
#endif
  MPID_HANDLE_MUTEX_LOCK( shandle );
  shandle->partner  = dest;
  MPID_HANDLE_MUTEX_UNLOCK( shandle ); 
  args              = NEW( thread_args );
  args->buf         = buf; 
  args->len         = len;
  args->src_lrank   = src_lrank;
  args->tag         = tag;
  args->context_id  = context_id;
  args->dest        = dest;
  args->msgrep      = msgrep;
  args->shandle     = shandle; 
  args->channel     = channel;
  
  marcel_attr_init(&attributes);	 
  marcel_attr_setdetachstate(&attributes, TRUE); 
  marcel_create( NULL, &attributes,
		 (marcel_func_t)MPID_CH_Mad_eager_tisend, args);
  
#ifdef GLOBAL_DEBUG
  marcel_printf(" ***** Exiting  short isend  **** \n");
#endif
  return mpi_errno; 
}
/*------------------------------------------------------------------*/
int MPID_CH_Mad_eager_recv( MPIR_RHANDLE *rhandle,
			    int from_grank,
			    void *in_pkt )
{
   CH_MAD_SHORT_ARGS *args = (CH_MAD_SHORT_ARGS *)in_pkt;
   int          msglen     = args->pkt->len;
   int          err        = MPI_SUCCESS; 
   
#ifdef DEBUG_POLLING
   marcel_printf("Entering Eager Recv : msglen : %i \n", msglen );
   marcel_printf("Locking mutex@(%p) ... ", &(rhandle->recv_handle.mutex));
   MPID_HANDLE_MUTEX_LOCK( rhandle );
   marcel_printf("Done.\n");
#else /* DEBUG POLLING */
   MPID_HANDLE_MUTEX_LOCK( rhandle );
#endif /* DEBUG POLLING */
   
   rhandle->s.MPI_TAG    = args->pkt->tag;  
   rhandle->s.MPI_SOURCE = args->pkt->lrank;
   MPID_CHK_MSGLEN( rhandle, msglen, err ); 
   rhandle->s.count	 = msglen;
   rhandle->s.MPI_ERROR  = err;
   MPID_HANDLE_STATUS_READY( rhandle );  /* syncro point */

#ifdef OPT
   if(msglen <= OPT_BUF_SIZE)
     {
	MEMCPY( rhandle->buf, args->buffer ,msglen);
     }
   else
#else /* OPT */  
   if(msglen > 0)
#endif /* OPT */      
     {
#ifdef ENFORCE_ALIGNMENT
	int _msglen = tbx_aligned( msglen , MAD_LENGTH_ALIGNMENT);
#else /* ENFORCE_ALIGNMENT */
	int _msglen = msglen ;
#endif /* ENFORCE_ALIGNMENT */     
	
	mad_unpack ( args->connection, 
		     rhandle->buf, 
		     _msglen,  
		     mad_send_CHEAPER, 
		     mad_receive_CHEAPER); 
     }
   mad_end_unpacking( args->connection );
   
   /* MPID_PKT_LEN_SET( args->pkt, 
    msglen + sizeof(MPID_HEAD_PKT_T)); */
   
   if(rhandle->finish)
     {
	(*(rhandle->finish))( rhandle );
     }
   
#ifdef DEBUG_POLLING
   marcel_printf("UNLocking mutex@(%p) ... ", &(rhandle->recv_handle.mutex));
   MPID_HANDLE_MUTEX_UNLOCK( rhandle );  
   marcel_printf("Done.\n");
   marcel_printf("Signaling Ready ...");
   MPID_HANDLE_COMPLETE_READY( rhandle ); /* syncro point */
   marcel_printf("Done.\n"); 
   marcel_printf("Complete Ready for rhandle @ %p\n", rhandle );
   marcel_printf("Exiting Eager Recv \n");
#else /* DEBUG POLLING */
   MPID_HANDLE_MUTEX_UNLOCK( rhandle );  
   MPID_HANDLE_COMPLETE_READY( rhandle ); /* syncro point */
#endif /* DEBUG POLLING */
   
   MPID_BREAK_LOOP();
   return err;
}
/*-----------------------------------------------------------------*/
int MPID_CH_Mad_eager_unxrecv_start( MPIR_RHANDLE *rhandle,
				     void *in_runex )
{
  MPIR_RHANDLE *runex = (MPIR_RHANDLE *)in_runex;
  int          err    = 0;  
  int          msglen; 
  
#ifdef GLOBAL_DEBUG
  marcel_printf("Entering Unex start\n");
#endif

  MPID_HANDLE_MUTEX_LOCK( runex );
  MPID_HANDLE_MUTEX_LOCK( rhandle );
  msglen = runex->s.count;
  MPID_CHK_MSGLEN( rhandle, msglen, err);
  
  /* Copy the data from the local area and free that area */
  if (msglen > 0) 
    {
      MEMCPY( rhandle->buf, runex->start, msglen );
      /* tbx_free((runex->recv_handle).buffer_list, runex->start);*/
      FREE( runex->start ); 
    }
  
  MPID_DO_HETERO(rhandle->msgrep = runex->msgrep);
  rhandle->s           = runex->s;
  rhandle->s.count     = msglen;
  rhandle->s.MPI_ERROR = err;
  MPID_HANDLE_STATUS_READY( rhandle );   /* syncro point */ 
  rhandle->wait        = NULL;
  rhandle->test        = NULL;
  rhandle->push        = NULL;
  if (rhandle->finish)
    {
      (*(rhandle->finish))( rhandle );  
    }
  MPID_HANDLE_MUTEX_UNLOCK( rhandle );
  MPID_HANDLE_MUTEX_UNLOCK( runex ); 
  MPID_HANDLE_COMPLETE_READY( rhandle ); /*  syncro point */
  
#ifdef GLOBAL_DEBUG
  marcel_printf("Complete Ready for rhandle @ %p (Unex) \n", rhandle );
  /* PRINT_RHANDLE( rhandle );*/
#endif

  MPID_RecvFree( runex );   
  
  BROADCAST_LOOP();  

#ifdef GLOBAL_DEBUG
  marcel_printf("Exiting Unex start\n");
#endif
  
  return err;
}
/*-----------------------------------------------------------------*/
/* Save an unexpected message in rhandle */
int MPID_CH_Mad_eager_save( MPIR_RHANDLE *rhandle,
			    int from,
			    void *in_pkt,
			    p_mad_channel_t unused_channel)
{
  CH_MAD_SHORT_ARGS *args   = (CH_MAD_SHORT_ARGS *)in_pkt;
  int                msglen = args->pkt->len; 
  
#ifdef DEBUG_POLLING
  marcel_printf("Entering Eager Save\n");
  marcel_printf("Locking mutex@(%p) ... ", &(rhandle->recv_handle.mutex));
  MPID_HANDLE_MUTEX_LOCK( rhandle );
  marcel_printf("Done.\n");
#else /* DEBUG POLLING */
  MPID_HANDLE_MUTEX_LOCK( rhandle );
#endif /* DEBUG POLLING */
  
  rhandle->s.MPI_TAG    = args->pkt->tag;
  rhandle->s.MPI_SOURCE = args->pkt->lrank;
  rhandle->s.MPI_ERROR  = 0; /* MPI_SUCCESS */
  rhandle->s.count      = msglen;
  
#ifdef DEBUG_POLLING
  marcel_printf("Signaling Status ready ... ");
  MPID_HANDLE_STATUS_READY( rhandle ); /* syncro point */  
  marcel_printf("Done.\n");  
#else /* DEBUG POLLING */
  MPID_HANDLE_STATUS_READY( rhandle ); /* syncro point */  
#endif /* DEBUG POLLING */

  if (msglen > 0)
    { 
      rhandle->start                     = args->buffer; 
      /* (rhandle->recv_handle).buffer_list = args->fast_buffer;  */
      /* Need to save msgrep for heterogeneous systems */
      MPID_DO_HETERO(rhandle->msgrep = (MPID_Msgrep_t)(args->pkt->msgrep));
    }
  rhandle->push     = MPID_CH_Mad_eager_unxrecv_start;
  rhandle->send_id  = args->pkt->send_id;
  
#ifdef DEBUG_POLLING
  marcel_printf("Signaling push ready ... ");
  MPID_HANDLE_PUSH_READY( rhandle );   /* syncro point */
  marcel_printf("Done.\n");  
  marcel_printf("UNLocking mutex@(%p) ... ", &(rhandle->recv_handle.mutex));
  MPID_HANDLE_MUTEX_UNLOCK( rhandle );
  marcel_printf("Done.\n");
  marcel_printf("Exiting Eager Save\n");
#else /* DEBUG POLLING */
  MPID_HANDLE_PUSH_READY( rhandle );   /* syncro point */
  MPID_HANDLE_MUTEX_UNLOCK( rhandle );
#endif /* DEBUG POLLING */

  return 0;
}
/*------------------------------------------------------------------*/
void MPID_CH_Mad_eager_delete( MPID_Protocol *p )
{
  FREE( p );
}
/*-----------------------------------------------------------------*/
MPID_Protocol *MPID_CH_Mad_Eager_setup( void )
{
  MPID_Protocol *p = NEW(MPID_Protocol);
  
  if (!p) 
    {
      return NULL;
    }
  p->send	 = MPID_CH_Mad_eager_send;
  p->recv	 = MPID_CH_Mad_eager_recv;
  p->isend	 = MPID_CH_Mad_eager_isend;
  p->wait_send   = NULL;
  p->push_send   = NULL;
  p->cancel_send = NULL;
  p->irecv	 = NULL;
  p->wait_recv   = NULL;
  p->push_recv   = NULL;
  p->cancel_recv = NULL;
  p->do_ack      = NULL;
  p->unex        = MPID_CH_Mad_eager_save;
  p->delete      = MPID_CH_Mad_eager_delete;
  
  return p;
}















