#ifndef _SYNCRO_H
#define _SYNCRO_H

#include "sbcnst2.h" /* includes pm2_common.h, for the NEW macro definition*/

//#define DEBUG_POLLING 

/*#define DS_LOCK_DEBUG */
/*#define LOCK_DEBUG*/
/*#define GLOBAL_DEBUG */
/*#define FINISH_DEBUG */
/*#define COMPLETION_DEBUG */
/*#define TERM_DEBUG*/
/* ref count in recv_handle for handler */
/*#define USED */

int              nbthreads;
int              MPID_MAD_Channel_Number; 
int              MPID_MAD_Global_Channel_Number; 
marcel_t         main_thread;

marcel_cond_t    condition;	  /* used in the barrier */
marcel_mutex_t   barrier;	  /* used in the barrier */
marcel_cond_t    pending_cond;    /* for while with MPID_n_pending variable see adi2init.c */
marcel_mutex_t   pending;	  /* for MPID_n_pending   variable */
marcel_cond_t    cancel_cond;     /* for while with expect_cancel_ack  variable see adi2init.c */
marcel_mutex_t   cancel;	  /* for expect_cancel_ack variable */
marcel_mutex_t   mutex_loop;      /* for the waitany loop */  
marcel_cond_t    cond_loop;       /* for the waitany loop */  



#define INIT_SYNCRO() { \
                        nbthreads = 0; \
			marcel_cond_init(&pending_cond,NULL); \
                        marcel_mutex_init(&pending, NULL); \
                        marcel_cond_init(&condition, NULL); \
		        marcel_mutex_init(&barrier, NULL); \
                        marcel_cond_init(&cond_loop, NULL); \
			marcel_mutex_init(&mutex_loop,NULL); \
 			marcel_cond_init(&cancel_cond,NULL); \
                        marcel_mutex_init(&cancel, NULL); \
		        for(index = 0; index < MPID_MAD_Channel_Number ; index++) \
			    marcel_mutex_init(&(MPID_MAD_Channel_Table[index].send_ok), NULL);}
			   			    
/*********************************** CHANNEL ACCESS *************************************/      
#define LOCK_CHANNEL( mutex )   marcel_mutex_lock( mutex )
#define UNLOCK_CHANNEL( mutex ) marcel_mutex_unlock( mutex )
/*********************************** CHANNEL ACCESS *************************************/      
/*********************************** PENDING REQUESTS *********************************/     
#define WAIT_PENDING() marcel_cond_wait(&pending_cond,&pending)
#define BROADCAST_PENDING() marcel_cond_broadcast(&pending_cond)
#define LOCK_PENDING() marcel_mutex_lock(&pending)
#define UNLOCK_PENDING() marcel_mutex_unlock(&pending)
#define MPID_INCREMENT_PENDING()  { LOCK_PENDING(); \
                                    MPID_n_pending++; \
				    UNLOCK_PENDING();}
#define MPID_DECREMENT_PENDING()  { LOCK_PENDING(); \
                                    if(--MPID_n_pending == 0) BROADCAST_PENDING(); \
				    UNLOCK_PENDING();}
/*********************************** PENDING REQUESTS *********************************/     
/***********************************    CANCEL ACKs   *********************************/     
#define WAIT_CANCEL_ACK()          marcel_cond_wait(&cancel_cond,&cancel)
#define BROADCAST_CANCEL_ACK()     marcel_cond_broadcast(&cancel_cond)
#define LOCK_CANCEL_ACK()          marcel_mutex_lock(&cancel)
#define UNLOCK_CANCEL_ACK()        marcel_mutex_unlock(&cancel)
#define MPID_INCREMENT_CANCEL_ACK()  { LOCK_CANCEL_ACK(); \
                                       expect_cancel_ack++; \
				       UNLOCK_CANCEL_ACK();}
#define MPID_DECREMENT_CANCEL_ACK()  { LOCK_CANCEL_ACK(); \
                                       if(--expect_cancel_ack == 0) BROADCAST_CANCEL_ACK(); \
				       UNLOCK_CANCEL_ACK();}
/***********************************    CANCEL ACKs   *********************************/     
/*************************************** LOOP *****************************************/
#ifdef LOCK_DEBUG
#define MPID_LOCK_LOOP() { marcel_printf("*****>> WAITLOOP: before lock in %s @ line %i \n",__FILE__,__LINE__); \
                           marcel_mutex_lock(&mutex_loop); \
			   marcel_printf("<<***** WAITLOOP: After lock \n");}

#define MPID_UNLOCK_LOOP() { marcel_printf("*****>> WAITLOOP: before unlock in %s @ line %i \n", __FILE__,__LINE__); \
                             marcel_mutex_unlock(&mutex_loop); \
			     marcel_printf("<<***** WAITLOOP: After unlock \n");}
#else /* LOCK_DEBUG */
#define MPID_LOCK_LOOP()   marcel_mutex_lock(&mutex_loop)
#define MPID_UNLOCK_LOOP() marcel_mutex_unlock(&mutex_loop)
#endif /* LOCK_DEBUG */
#define MPID_WAIT4EVENT() marcel_cond_wait(&cond_loop,&mutex_loop) 
#define BROADCAST_LOOP()  marcel_cond_broadcast(&cond_loop)
#define MPID_BREAK_LOOP() {MPID_LOCK_LOOP(); BROADCAST_LOOP(); MPID_UNLOCK_LOOP();} 
/*************************************** LOOP ******************************************/

#define BARRIER() { marcel_mutex_lock(&barrier); \
                    if(++nbthreads < ( MPID_MAD_Channel_Number + 1)) \
                       marcel_cond_wait(&condition,&barrier); \
                    else \
                       marcel_cond_broadcast(&condition); \
                    marcel_mutex_unlock(&barrier);}


/* those must be consistent with the MPID_RNDV_T def, 
   i.e typedef MPID_RNDV_t* syncro ... see mpid.h */
/* conditions */

/******************************************** COND4PUSH  ***********************************/
#define MPID_HANDLE_COND_DECLARE_4PUSH marcel_cond_t cond4push
#define MPID_HANDLE_COND_INIT_4PUSH( a ) marcel_cond_init(&((((MPIR_COMMON *)(a))->recv_handle).cond4push),NULL)
#define MPID_HANDLE_COND_WAIT_4PUSH( a ) marcel_cond_wait(&((((MPIR_COMMON *)(a))->recv_handle).cond4push), \
                                                           &((((MPIR_COMMON *)(a))->recv_handle).mutex4push))
#define MPID_HANDLE_COND_SIGNAL_4PUSH( a ) marcel_cond_broadcast(&((((MPIR_COMMON *)(a))->recv_handle).cond4push))
/******************************************** COND4PUSH    ***********************************/
/******************************************** COND4STATUS  ***********************************/
#define MPID_HANDLE_COND_DECLARE_4STATUS marcel_cond_t cond4status
#define MPID_HANDLE_COND_INIT_4STATUS( a ) marcel_cond_init(&((((MPIR_COMMON *)(a))->recv_handle).cond4status),NULL)
#define MPID_HANDLE_COND_WAIT_4STATUS( a ) marcel_cond_wait(&((((MPIR_COMMON *)(a))->recv_handle).cond4status), \
                                                           &((((MPIR_COMMON *)(a))->recv_handle).mutex4status))
#define MPID_HANDLE_COND_SIGNAL_4STATUS( a ) marcel_cond_broadcast(&((((MPIR_COMMON *)(a))->recv_handle).cond4status))
/******************************************** COND4STATUS    ***********************************/
/******************************************** COND4COMPLETE  ***********************************/
#define MPID_HANDLE_COND_DECLARE_4COMPLETE marcel_cond_t cond4complete
#define MPID_HANDLE_COND_INIT_4COMPLETE( a ) marcel_cond_init(&((((MPIR_COMMON *)(a))->recv_handle).cond4complete),NULL)
#define MPID_HANDLE_COND_WAIT_4COMPLETE( a ) marcel_cond_wait(&((((MPIR_COMMON *)(a))->recv_handle).cond4complete), \
							      &((((MPIR_COMMON *)(a))->recv_handle).mutex4complete))
#define MPID_HANDLE_COND_SIGNAL_4COMPLETE( a ) marcel_cond_broadcast(&((((MPIR_COMMON *)(a))->recv_handle).cond4complete))
/******************************************** COND4COMPLETE  ***********************************/
/******************************************** COND4CANCEL    ***********************************/
#define MPID_HANDLE_COND_DECLARE_4CANCEL marcel_cond_t cond4cancel
#define MPID_HANDLE_COND_INIT_4CANCEL( a ) marcel_cond_init(&((((MPIR_COMMON *)(a))->recv_handle).cond4cancel),NULL)
#define MPID_HANDLE_COND_WAIT_4CANCEL( a ) marcel_cond_wait(&((((MPIR_COMMON *)(a))->recv_handle).cond4cancel), \
							      &((((MPIR_COMMON *)(a))->recv_handle).mutex4cancel))
#define MPID_HANDLE_COND_SIGNAL_4CANCEL( a ) marcel_cond_broadcast(&((((MPIR_COMMON *)(a))->recv_handle).cond4cancel))
/******************************************** COND4CANCEL    ***********************************/
/******************************************** COND4USED      ***********************************/
#ifdef USED
#define MPID_HANDLE_COND_DECLARE_4USED marcel_cond_t cond4used
#define MPID_HANDLE_COND_INIT_4USED( a ) marcel_cond_init(&((((MPIR_COMMON *)(a))->recv_handle).cond4used),NULL)
#define MPID_HANDLE_COND_WAIT_4USED( a ) marcel_cond_wait(&((((MPIR_COMMON *)(a))->recv_handle).cond4used), \
                                                           &((((MPIR_COMMON *)(a))->recv_handle).mutex4used))
#define MPID_HANDLE_COND_SIGNAL_4USED( a ) marcel_cond_broadcast(&((((MPIR_COMMON *)(a))->recv_handle).cond4used))
#endif
/******************************************** COND4USED      ***********************************/

/* mutexes */
#define MPID_HANDLE_MUTEX_DECLARE marcel_mutex_t mutex 
#define MPID_HANDLE_MUTEX_INIT( a ) marcel_mutex_init(&((((MPIR_COMMON *)(a))->recv_handle).mutex),NULL)

#ifdef LOCK_DEBUG
#define MPID_HANDLE_MUTEX_LOCK( a ) marcel_printf("before lock in %s %i (%p) (%p)\n", __FUNCTION__, __LINE__,(a), \
                                         &((((MPIR_COMMON *)(a))->recv_handle).mutex)); \
                                    marcel_mutex_lock(&((((MPIR_COMMON *)(a))->recv_handle).mutex)); \
				    marcel_printf("after lock in %s (%p) (%p)\n", __FUNCTION__, (a), \
					 &((((MPIR_COMMON *)(a))->recv_handle).mutex))
     
#define MPID_HANDLE_MUTEX_UNLOCK( a ) marcel_printf("before unlock in %s %i (%p) (%p)\n", __FUNCTION__, __LINE__,(a), \
                                           &((((MPIR_COMMON *)(a))->recv_handle).mutex)); \
                                      marcel_mutex_unlock(&((((MPIR_COMMON *)(a))->recv_handle).mutex)); \
                                      marcel_printf("after unlock in %s (%p) (%p)\n", __FUNCTION__, (a), \
                                           &((((MPIR_COMMON *)(a))->recv_handle).mutex))

#define MPID_HANDLE_MUTEX_LOCK2( a ) marcel_printf("before lock in %s %i (%p) (%p)\n", __FUNCTION__, __LINE__, (a), \
                                         &((((MPIR_COMMON *)(a))->recv_handle).mutex)), \
                                     marcel_mutex_lock(&((((MPIR_COMMON *)(a))->recv_handle).mutex)), \
				     marcel_printf("after lock in %s (%p) (%p)\n", __FUNCTION__, (a), \
					 &((((MPIR_COMMON *)(a))->recv_handle).mutex))
     
#define MPID_HANDLE_MUTEX_UNLOCK2( a ) marcel_printf("before unlock in %s %i (%p) (%p)\n", __FUNCTION__, __LINE__, (a), \
                                           &((((MPIR_COMMON *)(a))->recv_handle).mutex)), \
				      marcel_mutex_unlock(&((((MPIR_COMMON *)(a))->recv_handle).mutex)), \
                                       marcel_printf("after unlock in %s (%p) (%p)\n", __FUNCTION__, (a), \
                                           &((((MPIR_COMMON *)(a))->recv_handle).mutex))
#else /* LOCK_DEBUG */
#define MPID_HANDLE_MUTEX_LOCK( a ) marcel_mutex_lock(&((((MPIR_COMMON *)(a))->recv_handle).mutex))
     
     /*  if ( locked() > 0) \ */
     /* 					{ \ */
     /* 					  marcel_printf("====================> LOCK value : %i (%s %i)\n"\ */
     /* 							,locked(), __FILE__,__LINE__); \ */
     /* 					  abort(); \ */
     /* 					};}	 */							         
     
#define MPID_HANDLE_MUTEX_UNLOCK( a ) marcel_mutex_unlock(&((((MPIR_COMMON *)(a))->recv_handle).mutex))
     
     /* if ( locked() >  0) \ */
     /* 					{ \ */
     /* 					  marcel_printf("====================> LOCK value : %i (%s %i)\n"\ */
     /* 							,locked(), __FILE__,__LINE__); \ */
     /* 					  abort(); \ */
     /* 					};}			 */						 

#endif /*LOCK_DEBUG */

#ifdef USED
#undef MPID_HANDLE_MUTEX_LOCK
#undef MPID_HANDLE_MUTEX_UNLOCK
#ifdef LOCK_DEBUG
#define MPID_HANDLE_MUTEX_LOCK( a ) { marcel_printf("before lock in %s %i (%p) (%p)\n", __FUNCTION__, __LINE__, (a), \
                                          &((((MPIR_COMMON *)(a))->recv_handle).mutex)); \
                                      marcel_mutex_lock(&((((MPIR_COMMON *)(a))->recv_handle).mutex)); \
				      marcel_printf("after lock in %s (%p) (%p)\n", __FUNCTION__, (a), \
					  &((((MPIR_COMMON *)(a))->recv_handle).mutex)); \
                                      MPID_HANDLE_INCR_USED( a );}
     
#define MPID_HANDLE_MUTEX_UNLOCK( a ) { MPID_HANDLE_DECR_USED( a ); \
                                        marcel_printf("before unlock in %s %i (%p) (%p)\n", __FUNCTION__, __LINE__, (a), \
                                            &((((MPIR_COMMON *)(a))->recv_handle).mutex)); \
                                        marcel_mutex_unlock(&((((MPIR_COMMON *)(a))->recv_handle).mutex)); \
                                        marcel_printf("after unlock in %s (%p) (%p)\n", __FUNCTION__, (a), \
					    &((((MPIR_COMMON *)(a))->recv_handle).mutex));}

#define MPID_HANDLE_MUTEX_LOCK_INTEST( a ) marcel_printf("before lock in %s %i (%p) (%p)\n", __FUNCTION__, __LINE__, (a), \
                                               &((((MPIR_COMMON *)(a))->recv_handle).mutex)), \
                                           marcel_mutex_lock(&((((MPIR_COMMON *)(a))->recv_handle).mutex)), \
				           marcel_printf("after lock in %s (%p) (%p)\n", __FUNCTION__, (a), \
					   &((((MPIR_COMMON *)(a))->recv_handle).mutex)), \
                                           MPID_HANDLE_INCR_USED_INTEST( a )
     
#define MPID_HANDLE_MUTEX_UNLOCK_INTEST( a ) MPID_HANDLE_DECR_USED_INTEST( a ), \
                                             marcel_printf("before unlock in %s %i (%p) (%p)\n", __FUNCTION__, __LINE__, (a), \
                                                 &((((MPIR_COMMON *)(a))->recv_handle).mutex)), \
                                             marcel_mutex_unlock(&((((MPIR_COMMON *)(a))->recv_handle).mutex)), \
                                             marcel_printf("after unlock in %s (%p) (%p)\n", __FUNCTION__, (a), \
                                                 &((((MPIR_COMMON *)(a))->recv_handle).mutex))

#else /* LOCK_DEBUG */
#define MPID_HANDLE_MUTEX_LOCK( a ) { marcel_mutex_lock(&((((MPIR_COMMON *)(a))->recv_handle).mutex)); \
				      MPID_HANDLE_INCR_USED( a ); }
       
#define MPID_HANDLE_MUTEX_UNLOCK( a ) { MPID_HANDLE_DECR_USED( a ); \
                                        marcel_mutex_unlock(&((((MPIR_COMMON *)(a))->recv_handle).mutex));}

#define MPID_HANDLE_MUTEX_LOCK_INTEST( a )   marcel_mutex_lock(&((((MPIR_COMMON *)(a))->recv_handle).mutex)), \
				             MPID_HANDLE_INCR_USED_INTEST( a )

#define MPID_HANDLE_MUTEX_UNLOCK_INTEST( a ) MPID_HANDLE_DECR_USED_INTEST( a ), \
                                             marcel_mutex_unlock(&((((MPIR_COMMON *)(a))->recv_handle).mutex))
#endif /*LOCK_DEBUG */
#endif /*USED */

#define MPID_HANDLE_MUTEX_DECLARE_4PUSH marcel_mutex_t mutex4push
#define MPID_HANDLE_MUTEX_INIT_4PUSH( a ) marcel_mutex_init(&((((MPIR_COMMON *)(a))->recv_handle).mutex4push),NULL)
#define MPID_HANDLE_MUTEX_LOCK_4PUSH( a ) marcel_mutex_lock(&((((MPIR_COMMON *)(a))->recv_handle).mutex4push))
#define MPID_HANDLE_MUTEX_UNLOCK_4PUSH( a ) marcel_mutex_unlock(&((((MPIR_COMMON *)(a))->recv_handle).mutex4push))

#define MPID_HANDLE_MUTEX_DECLARE_4STATUS marcel_mutex_t mutex4status
#define MPID_HANDLE_MUTEX_INIT_4STATUS( a ) marcel_mutex_init(&((((MPIR_COMMON *)(a))->recv_handle).mutex4status),NULL)
#define MPID_HANDLE_MUTEX_LOCK_4STATUS( a ) marcel_mutex_lock(&((((MPIR_COMMON *)(a))->recv_handle).mutex4status))
#define MPID_HANDLE_MUTEX_UNLOCK_4STATUS( a ) marcel_mutex_unlock(&((((MPIR_COMMON *)(a))->recv_handle).mutex4status))

#ifdef USED
#define MPID_HANDLE_MUTEX_DECLARE_4USED marcel_mutex_t mutex4used
#define MPID_HANDLE_MUTEX_INIT_4USED( a ) marcel_mutex_init(&((((MPIR_COMMON *)(a))->recv_handle).mutex4used),NULL)
#define MPID_HANDLE_MUTEX_LOCK_4USED( a ) marcel_mutex_lock(&((((MPIR_COMMON *)(a))->recv_handle).mutex4used))
#define MPID_HANDLE_MUTEX_UNLOCK_4USED( a ) marcel_mutex_unlock(&((((MPIR_COMMON *)(a))->recv_handle).mutex4used))
#endif

#ifdef DEBUG_CANCEL
#define MPID_HANDLE_MUTEX_LOCK_4CANCEL( a ) \
                                      marcel_printf("4CANCEL: before lock in %s %i (%p) (%p)\n", __FUNCTION__, __LINE__, (a), \
                                          &((((MPIR_COMMON *)(a))->recv_handle).mutex4cancel)); \
                                      marcel_mutex_lock(&((((MPIR_COMMON *)(a))->recv_handle).mutex4cancel)); \
				      marcel_printf("4CANCEL: after lock in %s (%p) (%p)\n", __FUNCTION__, (a), \
					  &((((MPIR_COMMON *)(a))->recv_handle).mutex4cancel))
                                      
#define MPID_HANDLE_MUTEX_UNLOCK_4CANCEL( a ) \
                                        marcel_printf("4CANCEL: before unlock in %s %i (%p) (%p)\n", __FUNCTION__, __LINE__, (a), \
                                            &((((MPIR_COMMON *)(a))->recv_handle).mutex4cancel)); \
                                        marcel_mutex_unlock(&((((MPIR_COMMON *)(a))->recv_handle).mutex4cancel)); \
                                        marcel_printf("4CANCEL: after unlock in %s (%p) (%p)\n", __FUNCTION__, (a), \
					&((((MPIR_COMMON *)(a))->recv_handle).mutex4cancel))
#else /* DEBUG_CANCEL */
#define MPID_HANDLE_MUTEX_LOCK_4CANCEL( a ) marcel_mutex_lock(&((((MPIR_COMMON *)(a))->recv_handle).mutex4cancel))
#define MPID_HANDLE_MUTEX_UNLOCK_4CANCEL( a ) marcel_mutex_unlock(&((((MPIR_COMMON *)(a))->recv_handle).mutex4cancel))
#endif /* DEBUG_CANCEL */
#define MPID_HANDLE_MUTEX_DECLARE_4CANCEL marcel_mutex_t mutex4cancel
#define MPID_HANDLE_MUTEX_INIT_4CANCEL( a ) marcel_mutex_init(&((((MPIR_COMMON *)(a))->recv_handle).mutex4cancel),NULL)


#define MPID_HANDLE_MUTEX_DECLARE_4COMPLETE marcel_mutex_t mutex4complete
#define MPID_HANDLE_MUTEX_INIT_4COMPLETE( a ) marcel_mutex_init(&((((MPIR_COMMON *)(a))->recv_handle).mutex4complete),NULL)

#ifdef USED
#ifdef COMPLETION_DEBUG
#define MPID_HANDLE_MUTEX_LOCK_4COMPLETE( a ) \
                                      { marcel_printf("4COMP: before lock in %s %i (%p) (%p)\n", __FUNCTION__, __LINE__, (a), \
                                          &((((MPIR_COMMON *)(a))->recv_handle).mutex4complete)); \
                                      marcel_mutex_lock(&((((MPIR_COMMON *)(a))->recv_handle).mutex4complete)); \
				      marcel_printf("4COMP: after lock in %s (%p) (%p)\n", __FUNCTION__, (a), \
					  &((((MPIR_COMMON *)(a))->recv_handle).mutex4complete)); \
                                      MPID_HANDLE_INCR_USED( a );}
     
#define MPID_HANDLE_MUTEX_UNLOCK_4COMPLETE( a ) \
                                        { MPID_HANDLE_DECR_USED( a ); \
                                        marcel_printf("4COMP: before unlock in %s %i (%p) (%p)\n", __FUNCTION__, __LINE__, (a), \
                                            &((((MPIR_COMMON *)(a))->recv_handle).mutex4complete)); \
                                        marcel_mutex_unlock(&((((MPIR_COMMON *)(a))->recv_handle).mutex4complete)); \
                                        marcel_printf("4COMP: after unlock in %s (%p) (%p)\n", __FUNCTION__, (a), \
					&((((MPIR_COMMON *)(a))->recv_handle).mutex4complete));}
#else /*  COMPLETION_DEBUG */
#define MPID_HANDLE_MUTEX_LOCK_4COMPLETE( a ) \
 { marcel_mutex_lock(&((((MPIR_COMMON *)(a))->recv_handle).mutex4complete)); \
   MPID_HANDLE_INCR_USED( a );}

#define MPID_HANDLE_MUTEX_UNLOCK_4COMPLETE( a ) \
{ MPID_HANDLE_DECR_USED( a ) ; \
  marcel_mutex_unlock(&((((MPIR_COMMON *)(a))->recv_handle).mutex4complete));}


#endif /*  COMPLETION_DEBUG */
#else /* USED */
#ifdef COMPLETION_DEBUG
#define MPID_HANDLE_MUTEX_LOCK_4COMPLETE( a ) \
                                      { marcel_printf("4COMP: before lock in %s %i (%p) (%p)\n", __FUNCTION__, __LINE__, (a), \
                                          &((((MPIR_COMMON *)(a))->recv_handle).mutex4complete)); \
                                        marcel_mutex_lock(&((((MPIR_COMMON *)(a))->recv_handle).mutex4complete)); \
				        marcel_printf("4COMP: after lock in %s (%p) (%p)\n", __FUNCTION__, (a), \
					  &((((MPIR_COMMON *)(a))->recv_handle).mutex4complete));}
#define MPID_HANDLE_MUTEX_UNLOCK_4COMPLETE( a ) \
                                        { marcel_printf("4COMP: before unlock in %s %i (%p) (%p)\n", __FUNCTION__, __LINE__, (a), \
                                            &((((MPIR_COMMON *)(a))->recv_handle).mutex4complete)); \
                                          marcel_mutex_unlock(&((((MPIR_COMMON *)(a))->recv_handle).mutex4complete)); \
                                          marcel_printf("4COMP: after unlock in %s (%p) (%p)\n", __FUNCTION__, (a), \
					  &((((MPIR_COMMON *)(a))->recv_handle).mutex4complete));}
#else
#define MPID_HANDLE_MUTEX_LOCK_4COMPLETE( a ) marcel_mutex_lock(&((((MPIR_COMMON *)(a))->recv_handle).mutex4complete))
#define MPID_HANDLE_MUTEX_UNLOCK_4COMPLETE( a ) marcel_mutex_unlock(&((((MPIR_COMMON *)(a))->recv_handle).mutex4complete))
#endif /* COMPLETION_DEBUG */
#endif /*USED */

/* struct syncro */
typedef struct{
  volatile int push_set;
  volatile int status_set;
#ifdef USED
  volatile int is_used;
  MPID_HANDLE_COND_DECLARE_4USED; 
  MPID_HANDLE_MUTEX_DECLARE_4USED; 
#endif /* USED */
  MPID_HANDLE_COND_DECLARE_4PUSH;
  MPID_HANDLE_MUTEX_DECLARE_4PUSH;
  MPID_HANDLE_COND_DECLARE_4STATUS;
  MPID_HANDLE_MUTEX_DECLARE_4STATUS;
  MPID_HANDLE_COND_DECLARE_4COMPLETE;
  MPID_HANDLE_MUTEX_DECLARE_4COMPLETE;
  MPID_HANDLE_COND_DECLARE_4CANCEL;
  MPID_HANDLE_MUTEX_DECLARE_4CANCEL;
  MPID_HANDLE_MUTEX_DECLARE;
  /*  p_tbx_memory_t buffer_list; */
} CH_MAD_SSYNCRO_t;
 
/*(((MPIR_COMMON *)(a))->recv_handle) = NEW( CH_MAD_SSYNCRO_t ); \ */
/*****************************************************************************************/
#ifdef USED
#define MPID_HANDLE_SSYNCRO_INIT( a ) { \
                                       MPID_HANDLE_COND_INIT_4USED( a ); \
                                       MPID_HANDLE_MUTEX_INIT_4USED( a ); \
                                       MPID_HANDLE_MUTEX_LOCK_4USED( a ); \
                                       (((MPIR_COMMON *)(a))->recv_handle).is_used = 0; \
                                       MPID_HANDLE_MUTEX_UNLOCK_4USED( a ); \
                                       MPID_HANDLE_MUTEX_INIT( a ); \
				       MPID_HANDLE_MUTEX_LOCK( a ); \
				       MPID_HANDLE_COND_INIT_4PUSH( a ); \
                                       MPID_HANDLE_MUTEX_INIT_4PUSH( a ); \
                                       MPID_HANDLE_COND_INIT_4STATUS( a ); \
                                       MPID_HANDLE_MUTEX_INIT_4STATUS( a ); \
				       MPID_HANDLE_COND_INIT_4COMPLETE( a ); \
                                       MPID_HANDLE_MUTEX_INIT_4COMPLETE( a ); \
                                       (((MPIR_COMMON *)(a))->recv_handle).push_set = 0; \
				       (((MPIR_COMMON *)(a))->recv_handle).status_set = 0; \
				       (((MPIR_COMMON *)(a))->recv_handle).buffer_list = NULL; \
		                       ((MPIR_COMMON *)(a))->self_index = 0; \
				       ((MPIR_COMMON *)(a))->ref_count  = 1; \
				       MPID_HANDLE_MUTEX_UNLOCK( a );}
#else /* USED */
#define MPID_HANDLE_SSYNCRO_INIT( a ) { \
                                       MPID_HANDLE_MUTEX_INIT( a ); \
				       MPID_HANDLE_MUTEX_LOCK( a ); \
				       MPID_HANDLE_COND_INIT_4PUSH( a ); \
                                       MPID_HANDLE_MUTEX_INIT_4PUSH( a ); \
                                       MPID_HANDLE_COND_INIT_4STATUS( a ); \
                                       MPID_HANDLE_MUTEX_INIT_4STATUS( a ); \
				       MPID_HANDLE_COND_INIT_4COMPLETE( a ); \
                                       MPID_HANDLE_MUTEX_INIT_4COMPLETE( a ); \
				       MPID_HANDLE_COND_INIT_4CANCEL( a ); \
                                       MPID_HANDLE_MUTEX_INIT_4CANCEL( a ); \
               			       (((MPIR_COMMON *)(a))->recv_handle).push_set = 0; \
				       (((MPIR_COMMON *)(a))->recv_handle).status_set = 0; \
				       ((MPIR_COMMON *)(a))->self_index = 0; \
				       ((MPIR_COMMON *)(a))->ref_count  = 1; \
				       MPID_HANDLE_MUTEX_UNLOCK( a );}
#endif /* USED */
/*   (((MPIR_COMMON *)(a))->recv_handle).buffer_list = NULL; \*/

#define MPID_HANDLE_SSYNCRO_FREE( a )  {}
/*{ FREE( (((MPIR_COMMON *)(a))->recv_handle) );} */
                                     
 				       

/*****************************************************************************************/
#define MPID_HANDLE_PUSH_READY( a ) MPID_HANDLE_MUTEX_LOCK_4PUSH( a ); \
                                    (((MPIR_COMMON *)(a))->recv_handle).push_set = 1; \
				    MPID_HANDLE_COND_SIGNAL_4PUSH( a ); \
				    MPID_HANDLE_MUTEX_UNLOCK_4PUSH( a )
     
#define MPID_HANDLE_WAIT4PUSH( a ) MPID_HANDLE_MUTEX_LOCK_4PUSH( a ); \
                                   while(!(((MPIR_COMMON *)(a))->recv_handle).push_set) \
				   MPID_HANDLE_COND_WAIT_4PUSH( a ); \
				   MPID_HANDLE_MUTEX_UNLOCK_4PUSH( a )
     
/***************************************************************************************/
     
#define MPID_HANDLE_STATUS_READY( a ) MPID_HANDLE_MUTEX_LOCK_4STATUS( a ); \
                                      (((MPIR_COMMON *)(a))->recv_handle).status_set = 1; \
                                      MPID_HANDLE_COND_SIGNAL_4STATUS( a ); \
				      MPID_HANDLE_MUTEX_UNLOCK_4STATUS( a )

#ifdef GLOBAL_DEBUG
#define MPID_HANDLE_WAIT4STATUS( a ) { marcel_printf("Waiting for STATUS in %s %i \n", __FILE__, __LINE__); \
				       MPID_HANDLE_MUTEX_LOCK_4STATUS( a ); \
				       while(!(((MPIR_COMMON *)(a))->recv_handle).status_set) \
				         MPID_HANDLE_COND_WAIT_4STATUS( a ); \
				       MPID_HANDLE_MUTEX_UNLOCK_4STATUS( a ); \
				       marcel_printf("Waiting for STATUS DONE\n");}
#else /*  GLOBAL_DEBUG */
#define MPID_HANDLE_WAIT4STATUS( a ) { MPID_HANDLE_MUTEX_LOCK_4STATUS( a ); \
				       while(!(((MPIR_COMMON *)(a))->recv_handle).status_set) \
				         MPID_HANDLE_COND_WAIT_4STATUS( a ); \
				       MPID_HANDLE_MUTEX_UNLOCK_4STATUS( a );}
#endif /*  GLOBAL_DEBUG */
/*****************************************************************************************/
#ifdef USED
#define MPID_HANDLE_DECR_USED( a ) { MPID_HANDLE_MUTEX_LOCK_4USED( a ); \
                                     if ( (--((((MPIR_COMMON *)(a))->recv_handle).is_used)) == 0) \
				        MPID_HANDLE_COND_SIGNAL_4USED( a ); \
                                     MPID_HANDLE_MUTEX_UNLOCK_4USED( a );}

#define MPID_HANDLE_DECR_USED_INTEST( a ) \
(MPID_HANDLE_MUTEX_LOCK_4USED( a ), (--(((MPIR_COMMON *)(a))->recv_handle.is_used) == 0)) \
?(MPID_HANDLE_COND_SIGNAL_4USED( a ),MPID_HANDLE_MUTEX_UNLOCK_4USED( a ), 1) \
:(MPID_HANDLE_MUTEX_UNLOCK_4USED( a ),0)

     
#define MPID_HANDLE_INCR_USED( a ) { MPID_HANDLE_MUTEX_LOCK_4USED( a ); \
                                     ((((MPIR_COMMON *)(a))->recv_handle).is_used)++ ;\
				     MPID_HANDLE_MUTEX_UNLOCK_4USED( a );}

#define MPID_HANDLE_INCR_USED_INTEST( a ) MPID_HANDLE_MUTEX_LOCK_4USED( a ), \
                                          ((((MPIR_COMMON *)(a))->recv_handle).is_used)++ , \
				          MPID_HANDLE_MUTEX_UNLOCK_4USED( a )

#define MPID_HANDLE_WAIT4USED( a ) { MPID_HANDLE_MUTEX_LOCK_4USED( a ); \
                                     while( ((((MPIR_COMMON *)(a))->recv_handle).is_used) != 0) \
				       MPID_HANDLE_COND_WAIT_4USED( a ); \
				     MPID_HANDLE_MUTEX_UNLOCK_4USED( a );}
#else /* USED */
#define MPID_HANDLE_WAIT4USED( a ) {}
#endif /* USED */
/*****************************************************************************************/
#define MPID_HANDLE_COMPLETE_READY( a ) { MPID_HANDLE_MUTEX_LOCK_4COMPLETE( a ); \
					  ((MPIR_COMMON *)(a))->is_complete = 1; \
                                          MPID_HANDLE_COND_SIGNAL_4COMPLETE( a ); \
                                          MPID_HANDLE_MUTEX_UNLOCK_4COMPLETE( a );}

#define MPID_HANDLE_WAIT4COMPLETE( a ) { MPID_HANDLE_MUTEX_LOCK_4COMPLETE( a ); \
                                         while(!((MPIR_COMMON *)(a))->is_complete) \
				           MPID_HANDLE_COND_WAIT_4COMPLETE( a ); \
					 MPID_HANDLE_MUTEX_UNLOCK_4COMPLETE( a );}

#define MPID_HANDLE_COMPLETE_RESET( a ) { MPID_HANDLE_MUTEX_LOCK_4COMPLETE( a ); \
                                          ((MPIR_COMMON *)(a))->is_complete = 0; \
					  MPID_HANDLE_MUTEX_UNLOCK_4COMPLETE( a );}
          
#define MPID_HANDLE_TEST_4COMPLETE( a ) (((((MPIR_COMMON *)(a))->is_complete) == 1 ) ? 1:0 )
/*****************************************************************************************/
#define MPID_HANDLE_CANCEL_READY( a ) { MPID_HANDLE_MUTEX_LOCK_4CANCEL( a ); \
					((MPIR_SHANDLE *)(a))->cancel_complete = 1; \
                                        MPID_HANDLE_COND_SIGNAL_4CANCEL( a ); \
                                        MPID_HANDLE_MUTEX_UNLOCK_4CANCEL( a );}

#define MPID_HANDLE_WAIT4CANCEL( a ) { MPID_HANDLE_MUTEX_LOCK_4CANCEL( a ); \
                                       while(!(((MPIR_SHANDLE *)(a))->cancel_complete)) \
				         MPID_HANDLE_COND_WAIT_4CANCEL( a ); \
				       MPID_HANDLE_MUTEX_UNLOCK_4CANCEL( a );}

#define MPID_HANDLE_CANCEL_RESET( a ) { MPID_HANDLE_MUTEX_LOCK_4CANCEL( a ); \
                                        ((MPIR_SHANDLE *)(a))->cancel_complete = 0; \
					MPID_HANDLE_MUTEX_UNLOCK_4CANCEL( a );}
/*****************************************************************************************/
#ifdef LOCK_DEBUG
#define MPID_HANDLE_TEST_FUNC( a , func ) (MPID_HANDLE_MUTEX_LOCK2( a ), (a)->func ) \
                                          ?(MPID_HANDLE_MUTEX_UNLOCK2( a ), 1) \
					  :(MPID_HANDLE_MUTEX_UNLOCK2( a ), 0)
#else /* LOCK_DEBUG */
#define MPID_HANDLE_TEST_FUNC( a , func ) (MPID_HANDLE_MUTEX_LOCK( a ), (a)->func ) \
                                          ?(MPID_HANDLE_MUTEX_UNLOCK( a ), 1) \
					  :(MPID_HANDLE_MUTEX_UNLOCK( a ), 0)
#endif /* LOCK_DEBUG */

#ifdef USED
#undef MPID_HANDLE_TEST_FUNC
#define MPID_HANDLE_TEST_FUNC( a , func ) (MPID_HANDLE_MUTEX_LOCK_INTEST( a ), (a)->func ) \
                                          ?(MPID_HANDLE_MUTEX_UNLOCK_INTEST( a ), 1) \
					  :(MPID_HANDLE_MUTEX_UNLOCK_INTEST( a ), 0)
#endif /*USED */

/******************************************************************************/

/* those are used in both mpid.h and  sbcnst2.c 
   the definitions are now centralized ... */
/*
 * Thread definitions.  We show an example of pthreads, as well as
 * a default set for no threading.  
 */
#if defined(HAVE_PTHREAD_MUTEX_INIT) && defined(USE_PTHREADS)
#define MPID_THREAD_DS_LOCK_DECLARE pthread_mutex_t mutex
#define MPID_THREAD_DS_LOCK_INIT(p) pthread_mutex_init( &(p)->mutex,pthread_mutexattr_default )
#define MPID_THREAD_DS_LOCK(p)      pthread_mutex_lock( &(p)->mutex )
#define MPID_THREAD_DS_UNLOCK(p)    pthread_mutex_unlock( &(p)->mutex )
#define MPID_THREAD_DS_LOCK_FREE(p) pthread_mutex_destroy( &(p)->mutex )
#else
#define MPID_THREAD_DS_LOCK_DECLARE marcel_mutex_t mutex
#define MPID_THREAD_DS_LOCK_INIT(p) marcel_mutex_init( &((p)->mutex),NULL)

/*#define DS_LOCK_DEBUG*/
#ifdef DS_LOCK_DEBUG
#define MPID_THREAD_DS_LOCK( p ) marcel_printf("before lock in %s mutex@(%p)\n", __FUNCTION__,&((p)->mutex)); \
                                 marcel_mutex_lock( &((p)->mutex) ); \
				 marcel_printf("after lock in %s mutex@(%p)\n", __FUNCTION__,&((p)->mutex))
     
#define MPID_THREAD_DS_UNLOCK( p ) marcel_printf("before unlock in %s mutex@(%p)\n", __FUNCTION__,&((p)->mutex)); \
                                   marcel_mutex_unlock( &((p)->mutex) ); \
				   marcel_printf("after unlock in %s mutex@(%p)\n", __FUNCTION__,&((p)->mutex))
#else  /* DS_LOCK_DEBUG */
#define MPID_THREAD_DS_LOCK(p) marcel_mutex_lock( &((p)->mutex) )

     /* if ( locked() > 0) \ */
     /*                                 { \ */
     /* 				    marcel_printf("====================> LOCK value : %i (%s %i)\n"\ */
     /* 						  ,locked(), __FILE__,__LINE__); \ */
     /* 				    abort(); \ */
     /* 			        } */
     
#define MPID_THREAD_DS_UNLOCK(p)  marcel_mutex_unlock( &((p)->mutex) )

                                  /* if ( locked() > 0) \ */
     /*                                    { \ */
     /* 				     marcel_printf("====================> LOCK value : %i (%s %i)\n"\ */
     /* 				     ,locked(), __FILE__,__LINE__); \ */
     /* 				     abort(); \ */
     /* 			            } */

#endif /* DS_LOCK_DEBUG */

#define MPID_THREAD_DS_TRYLOCK( p )        marcel_mutex_trylock( &((p)->mutex) )
#define MPID_THREAD_DS_LOCK_FREE( p )      marcel_mutex_destroy( &((p)->mutex) )
#define MPID_THREAD_SELF()                 marcel_self()
#define MPID_THREAD_DS_COND_DECLARE        marcel_cond_t condition
#define MPID_THREAD_DS_COND_INIT( p )      marcel_cond_init( &(p)->condition, NULL )
#define MPID_THREAD_DS_COND_SIGNAL( p )    marcel_cond_signal( &(p)->condition )
#define MPID_THREAD_DS_COND_BROADCAST( p ) marcel_cond_broadcast( &(p)->condition )
#define MPID_THREAD_DS_COND_FREE( p )      marcel_cond_free( &(p)->condition )
#endif

#endif /* SYNCRO_H*/











