#!/bin/sh

# PM2: Parallel Multithreaded Machine
# Copyright (C) 2001 "the PM2 team" (see AUTHORS file)
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.


PROGNAME=pm2-gen-make.sh

#############################################################
# Setting destination directory

make_dir() {
    ldir=$1

    if [ ! -d $ldir ]; then
#	echo "Creating directory : $ldir"
	mkdir -p $ldir
    fi
}

############################################################
# Generating Makefiles

## Libs ########################################

gen_libs() {

LIBRARY=$1
dst=$2

echo Generating $dst

eval _PM2CONFIG_LIB_DIR='$'_PM2CONFIG_${LIBRARY}_LIB_DIR
eval _PM2CONFIG_OBJ_DIR='$'_PM2CONFIG_${LIBRARY}_OBJ_DIR
eval _PM2CONFIG_CPP_DIR='$'_PM2CONFIG_${LIBRARY}_CPP_DIR
eval _PM2CONFIG_ASM_DIR='$'_PM2CONFIG_${LIBRARY}_ASM_DIR
eval _PM2CONFIG_DEP_DIR='$'_PM2CONFIG_${LIBRARY}_DEP_DIR
eval _PM2CONFIG_BIN_DIR='$'_PM2CONFIG_${LIBRARY}_BIN_DIR
eval _PM2CONFIG_TMP_DIR='$'_PM2CONFIG_${LIBRARY}_TMP_DIR
eval _PM2CONFIG_SRC_DIR='$'_PM2CONFIG_${LIBRARY}_SRC_DIR
eval _PM2CONFIG_INCLUDE_DIR='$'_PM2CONFIG_${LIBRARY}_INCLUDE_DIR
# Nothing to do for stampdir...
# Nothing to do for (flavor) stampfile
# Nothing to do for extension
eval _PM2CONFIG_KERNEL_CFLAGS='$'_PM2CONFIG_${LIBRARY}_KERNEL_CFLAGS
# Nothing to do for CC
eval _PM2CONFIG_BUILD_DYNAMIC='$'_PM2CONFIG_${LIBRARY}_BUILD_DYNAMIC
eval _PM2CONFIG_BUILD_STATIC='$'_PM2CONFIG_${LIBRARY}_BUILD_STATIC

cat > $dst <<EOF
# Do not edit ! Generated by $PROGNAME

LIB_GEN_LIB := $_PM2CONFIG_LIB_DIR
LIB_GEN_OBJ := $_PM2CONFIG_OBJ_DIR
LIB_GEN_CPP := $_PM2CONFIG_CPP_DIR
LIB_GEN_ASM := $_PM2CONFIG_ASM_DIR
LIB_GEN_DEP := $_PM2CONFIG_DEP_DIR
LIB_GEN_BIN := $_PM2CONFIG_BIN_DIR
LIB_GEN_TMP := $_PM2CONFIG_TMP_DIR
LIB_GEN_SRC := $_PM2CONFIG_SRC_DIR
LIB_GEN_INC := $_PM2CONFIG_INCLUDE_DIR
LIB_GEN_STAMP := $_PM2CONFIG_STAMP_DIR
LIB_STAMP_FLAVOR := $_PM2CONFIG_FLAVOR_STAMP_FILE
LIB_EXT := $_PM2CONFIG_EXT

LIB_CFLAGS := $_PM2CONFIG_KERNEL_CFLAGS
CFLAGS += \$(LIB_CFLAGS)

CC := $_PM2CONFIG_CC

BUILD_DYNAMIC_LIBS := $_PM2CONFIG_BUILD_DYNAMIC
BUILD_STATIC_LIBS := $_PM2CONFIG_BUILD_STATIC
EOF

}


## Progs ########################################

gen_progs() {

dst=$1

if [ -z "$PROGRAM" ]; then
    echo "Error : variable PROGRAM not set !"
    exit -1
fi

echo Generating $dst

eval _PM2CONFIG_BIN_DIR='$'_PM2CONFIG_${PROGRAM}_BIN_DIR
eval _PM2CONFIG_OBJ_DIR='$'_PM2CONFIG_${PROGRAM}_OBJ_DIR
eval _PM2CONFIG_CPP_DIR='$'_PM2CONFIG_${PROGRAM}_CPP_DIR
eval _PM2CONFIG_ASM_DIR='$'_PM2CONFIG_${PROGRAM}_ASM_DIR
eval _PM2CONFIG_DEP_DIR='$'_PM2CONFIG_${PROGRAM}_DEP_DIR
eval _PM2CONFIG_SRC_DIR='$'_PM2CONFIG_${PROGRAM}_SRC_DIR
eval _PM2CONFIG_INCLUDE_DIR='$'_PM2CONFIG_${PROGRAM}_INCLUDE_DIR
# Nothing to do for stampdir...
# Nothing to do for (flavor) stampfile
# Nothing to do for extension
eval _PM2CONFIG_KERNEL_CFLAGS='$'_PM2CONFIG_${PROGRAM}_KERNEL_CFLAGS
# Nothing to do for libs
# Nothing to do for libs-l
# Nothing to do for build dir
# Nothing to do for CC

cat > $dst <<EOF
# Do not edit ! Generated by $PROGNAME

PRG_GEN_BIN := $_PM2CONFIG_BIN_DIR
PRG_GEN_OBJ := $_PM2CONFIG_OBJ_DIR
PRG_GEN_CPP := $_PM2CONFIG_CPP_DIR
PRG_GEN_ASM := $_PM2CONFIG_ASM_DIR
PRG_GEN_DEP := $_PM2CONFIG_DEP_DIR
PRG_GEN_SRC := $_PM2CONFIG_SRC_DIR
PRG_GEN_INC := $_PM2CONFIG_INCLUDE_DIR
PRG_GEN_STAMP := $_PM2CONFIG_STAMP_DIR
PRG_STAMP_FLAVOR := $_PM2CONFIG_FLAVOR_STAMP_FILE
PRG_STAMP_FILES := $_PM2CONFIG_LIB_STAMP_FILES
PRG_EXT := $_PM2CONFIG_EXT

PRG_CFLAGS := $_PM2CONFIG_KERNEL_CFLAGS
CFLAGS += \$(PRG_CFLAGS) -I\$(PRG_GEN_INC)
                        #^^^^^^^^^^^^^^^ should be included in pm2-config

PRG_LDFLAGS := $_PM2CONFIG_LIBS
LDFLAGS += \$(PRG_LDFLAGS)

PRG_LLIBS += $_PM2CONFIG_LD_LIBS

# Necessaire ?
PRG_BUILDDIR :=  $_PM2CONFIG_BUILD

CC := $_PM2CONFIG_CC

EOF
}

## Apps ########################################

gen_apps() {

dst=$1

echo Generating $dst

cat > $dst <<EOF
# Do not edit ! Generated by $PROGNAME

LIB_GEN_STAMP := $_PM2CONFIG_STAMP_DIR
CC := $_PM2CONFIG_CC

APP_EXT := $_PM2CONFIG_EXT
APP_LOADER := $_PM2CONFIG_LOADER
APP_CFLAGS += $_PM2CONFIG_appli_KERNEL_CFLAGS
APP_LDFLAGS += $_PM2CONFIG_LIBS
APP_LLIBS += $_PM2CONFIG_LD_LIBS
APP_STAMP_FLAVOR := $_PM2CONFIG_FLAVOR_STAMP_FILE
APP_STAMP_FILES := $_PM2CONFIG_LIB_STAMP_FILES
APP_BUILDD := $_PM2CONFIG_BUILD

CFLAGS += \$(APP_CFLAGS)
LDFLAGS += \$(APP_LDFLAGS)

EOF

case " $_PM2CONFIG_CFLAGS " in
    *\ -DAPPLI_PROFILE\ *)
	echo "APP_PROFILE := yes" >> $dst
	;;
    *)
	echo "APP_PROFILE := no" >> $dst
	;;
esac

}

############################################################
# Main

PM2_FLAVOR=$1

if [ $# -gt 1 ]; then
    target=$2
else
    target=all
fi

set -- --source-mode
. ${PM2_ROOT}/bin/pm2-config

# Generating directories

dir=$_PM2CONFIG_BUILD
LIBRARIES="$_PM2CONFIG_MODULES"
destdir=$_PM2CONFIG_MAK_DIR

# Generating .mak for libraries

make_dir $destdir

if [ $target = dir ]; then
    exit 0
fi

if [ $target = progs ]; then 
    gen_progs $destdir/progs-config.mak
    exit 0
fi

if [ $target = apps ]; then
    gen_apps $destdir/apps-config.mak
    exit 0
fi

if [ $target = all -o $target = main-config ]; then
    dst="$destdir/main-config.mak"
    echo Generating $dst
    echo "# Do not edit ! Generated by $PROGNAME" > $dst
    echo "" >> $dst
    echo "MAIN_STAMP_FLAVOR := $_PM2CONFIG_FLAVOR_STAMP_FILE" >> $dst
    echo "LIBS := $LIBRARIES" >> $dst
    echo "PROF_LIBS := $_PM2CONFIG_PROF_MODULES" >> $dst
    echo "DOT_H_GEN_LIBS := $_PM2CONFIG_HEADGEN_MODULES" >> $dst
    case " $_PM2CONFIG_CFLAGS " in
	*\ -DAPPLI_PROFILE\ *)
	    echo "APP_PROFILE := yes" >> $dst
	    ;;
	*)
	    echo "APP_PROFILE := no" >> $dst
	    ;;
    esac
fi

for i in $LIBRARIES; do
    if [ $target = all -o $target = $i ]; then
	gen_libs $i $destdir/$i-config.mak
    fi
done

exit 0
